#include "Otr/OtrI.h"
#include "Otr/OtrAlign.h"
#include "Otr/OtrDb.h"

#include <stdio.h>
#include <stdlib.h>

/* Fortran Interface */
#ifdef __cplusplus
extern "C" {
#endif
  void  OtrAlignSwitch_(int* i)        { OtrAlignSwitch(*i); }
  int   OtrAlignInitialize_()          { return OtrAlignInitialize(); }
  int   OtrAlignReinitialize_()        { return OtrAlignUpdate(); }
  int   OtrAlignWriteToDb_()           { return OtrAlignWriteToDb(); }
  int   OtrAlignClose_()               { return OtrAlignClose(); }
  int   OtrAlignScan_()                { return OtrAlignScan(0); }
  int   OtrAlignPrint_()               { return OtrAlignPrint(0); }
  int   OtrAlignTestMe_()              { return OtrAlignTestMe(); }
#ifdef __cplusplus
  }
#endif
  
/*------------------- Structures -------------------------- */
/* Structure of OTr Alignment database                      */
typedef struct {
  char hwc[12];      /* GEDE name: hwc = -t2-3-40-01 = sign-SL-lay-mod-sec */
  float ang[3];      /* three rotations */ 
  float eang[3];     /* error for rotations */
  float delt[3];     /* three translations */
  float edelt[3];    /* error for translations */
  
  /* OTR specific stuff */
  float sloff[3];    /* offset of sublayer (=monolayer) within module */
  float esloff[3];   /* error of sublayer offset */
  float redres;      /* width of residual distribution 
                        extrapolated to a track momentum of infinity */
} OtrAlignStruct;


/* ----------------------------------------------------------------------*/
OtrDbField structfmtAlign[]={
  {"char[]", "hwc",     OtrDbOffset(OtrAlignStruct,hwc)},
  {"float",  "ang0",    OtrDbOffset(OtrAlignStruct,ang[0])},
  {"float",  "ang1",    OtrDbOffset(OtrAlignStruct,ang[1])},
  {"float",  "ang2",    OtrDbOffset(OtrAlignStruct,ang[2])},
  {"float",  "eang0",   OtrDbOffset(OtrAlignStruct,eang[0])},
  {"float",  "eang1",   OtrDbOffset(OtrAlignStruct,eang[1])},
  {"float",  "eang2",   OtrDbOffset(OtrAlignStruct,eang[2])},
  {"float",  "delt0",   OtrDbOffset(OtrAlignStruct,delt[0])},
  {"float",  "delt1",   OtrDbOffset(OtrAlignStruct,delt[1])},
  {"float",  "delt2",   OtrDbOffset(OtrAlignStruct,delt[2])},
  {"float",  "edelt0",  OtrDbOffset(OtrAlignStruct,edelt[0])},
  {"float",  "edelt1",  OtrDbOffset(OtrAlignStruct,edelt[1])},
  {"float",  "edelt2",  OtrDbOffset(OtrAlignStruct,edelt[2])},
  {"float",  "sloff0",  OtrDbOffset(OtrAlignStruct,sloff[0])},
  {"float",  "sloff1",  OtrDbOffset(OtrAlignStruct,sloff[1])},
  {"float",  "sloff2",  OtrDbOffset(OtrAlignStruct,sloff[2])},
  {"float",  "esloff0", OtrDbOffset(OtrAlignStruct,esloff[0])},
  {"float",  "esloff1", OtrDbOffset(OtrAlignStruct,esloff[1])},
  {"float",  "esloff2", OtrDbOffset(OtrAlignStruct,esloff[2])},
  {"float",  "redres",  OtrDbOffset(OtrAlignStruct,redres)}
};

static OtrDbTable AlignTableDef = {"OTR_Align",1,"OTR Align table",sizeof(OtrAlignStruct),
				   structfmtAlign,sizeof(structfmtAlign)/sizeof(OtrDbField)};


/* ----------------------------------------------------------------------*/

void OtrAlignDefine() {
  OtrDbDefine("/OTR_ALIGN",AlignTableDef.name,OtrAlignUpdate);
  }


void OtrAlignUserDefine(char *server,int maj,int min,int mat,int ver) {
  OtrDbUserDefine(server,AlignTableDef.name,OtrAlignUpdate,
               maj,min,mat,ver);
  }

/* ----------------------------------------------------------------------*/


/* Structure for Alignment parameters */
typedef struct {
  float ra[9];      /* matrix of alignment rotation (in Laboratory frame) */
  float ta[3];      /* alignment translation shift (in NEDE frame) */
} OtrAlignParStruct;

/** @name OtrAlign

    The OtrAlign interface provides access to the alignment rotation of
    3-vectors. 
 */
/*@{*/

/*------------------- Static variables -------------------------- */
/** Algorithm used for OTR alignment. */
//static int OtrAlignAlgorithm=OTRIInvalid;
static int OtrAlignAlgorithm=0;

/** Number of Alignment structures filled
    (usually coincides with OtrGeoNumberOfVolumes()).
    */
static int numberofalign=0;

/** Pointer to Alignment data - copy of Database structure.
 */
static OtrAlignStruct* aligns = 0;

/**
 Pointer to Alignment parameters -
 additional parameters for fast alignment calculations.
 */
static OtrAlignParStruct* alignpars = 0;


//#define DEBUG 1

/*------------------- Functions -------------------------- */
/** Switch between different solutions of the task of the package.

  @param   i  integer number (currently any number except OTRIInvalid)
  */
void  OtrAlignSwitch(int i) {
  OtrAlignAlgorithm=i;
  }

/** Initialize OTrAlign package.

  All function of the package (beside OtrAlignSetAlgorithm) check
  whether the initialization was done properly.

  @return 0 if state ok, else error code
  */
  
int   OtrAlignInitialize() {
  if (aligns!=0) return OtrIValid();
  if (OtrAlignAlgorithm==OTRIInvalid) OtrAlignAlgorithm=0;
  return OtrAlignUpdate();
  }



#ifdef USE_ARTE
/* Copy ARTE NEDE table to GEDE table and apply alignment to the coordinates.
   (Only internal usage)
 */
void UpdateGede(int gi, float Grot[9], float trans[3]) {
    int inede=OtrGeoGet(gi).index();
    ArtePointer<GEDE> g(inede);
    ArtePointer<NEDE> n(inede);
    
    memcpy(&(g->cmp), &(n->cmp), sizeof(NEDE));

    for(int i=0;i<9;i++) {
      g->rot[i] = Grot[i];
    };

    float deltax = trans[0];
    float deltay = trans[1];
    float deltaz = trans[2];

    for(int i=0; i<n->nsl; i++) {
      g->xfw[i] = n->xfw[i] + deltax;
      g->zfw[i] = n->zfw[i] + deltaz;
      g->y1w[i] = n->y1w[i] + deltay;
      g->y2w[i] = n->y2w[i] + deltay;
    }

    g->xmin = n->xmin + deltax;
    g->xmax = n->xmax + deltax;
    	       
    g->ymin = n->ymin + deltay;
    g->ymax = n->ymax + deltay;
    	       
    g->zmin = n->zmin + deltaz;
    g->zmax = n->zmax + deltaz;
    	       
    g->sxmn = n->sxmn + deltax;
    g->sxmx = n->sxmx + deltax;
    	       
    g->symn = n->symn + deltay;
    g->symx = n->symx + deltay;
}
#endif

/* Fill OtrAlignStruct for given gi
   (Only internal usage).
   */
int OtrAlignFill(int gi,
		 float dx, float dy, float dz,
		 float a, float b, float c) {

  OtrAlignStruct *al = &aligns[gi];
  strncpy(al->hwc,OtrGeoGet(gi)->hwc,12);

  al->ang[0] = a;
  al->ang[1] = b;
  al->ang[2] = c;
  al->eang[0] = 1.e-5;
  al->eang[1] = 1.e-5;
  al->eang[2] = 1.e-5;
  
  al->delt[0] = dx;
  al->delt[1] = dy;
  al->delt[2] = dz;
  al->edelt[0] = 0.01;
  al->edelt[1] = 0.01;
  al->edelt[2] = 0.01;
  
  al->sloff[0] = 0.;
  al->sloff[1] = 0.;
  al->sloff[2] = 0.;
  al->esloff[0] = 0.01;
  al->esloff[1] = 0.01;
  al->esloff[2] = 0.01;
  
  al->redres = 0.;

  return OTRIValid;
}  

/* Calculate OtrAlignParStruct for given gi
   (Only internal usage).
   */
int OtrAlignParCalc(int gi) {

#ifdef USE_ARTE
  ArtePointer<NEDE> n= OtrGeoGet(gi);
#else
  NEDE* n= (NEDE*)OtrGeoGet(gi);
#endif

  OtrAlignStruct *al = &aligns[gi];
  OtrAlignParStruct *alpar = &alignpars[gi];
  
  char buf[100];
  float delt[3];
  for (int i=0; i<3; i++) {
    if (fabs(al->ang[i])>0.1) {
      sprintf(buf,"a = %f rad",al->ang[i]);
      return OtrIErrorMsg(OTRIInvalid,"OtrAlignParCalc",
			  "Alignment angle too large:",buf);
    }

    // In Dirk's alignment files delta shifts are meant
    //  to be subtracted from absolute position, not added.
    //
    delt[i]= -al->delt[i];
    if (fabs(al->delt[i])>20.) {
      sprintf(buf,"dx = %f cm",al->delt[i]);
      return OtrIErrorMsg(OTRIInvalid,"OtrAlignParCalc",
			  "Alignment shift too large:",buf);
    }
  }

  float a = al->ang[0];
  float b = al->ang[1];
  float c = al->ang[2];
  
  /* rotation matrix RA:=R_z(c)*R_y(b)*R_x(a)= */
  
  float RA[9]=
 {cos(b)*cos(c)                     ,  cos(b)*sin(c)                     , -sin(b),
 -cos(a)*sin(c)+sin(a)*sin(b)*cos(c),  cos(a)*cos(c)+sin(a)*sin(b)*sin(c), sin(a)*cos(b),
  sin(a)*sin(c)+cos(a)*sin(b)*cos(c), -sin(a)*cos(c)+cos(a)*sin(b)*sin(c), cos(a)*cos(b)
 };

  int i,k;
  for(i=0; i<9; i++) alpar->ra[i] = RA[i];
  
  float cog[3],N[9];
  /* c.o.g. for nede */
  if (n->sec != 0) {
    cog[0]=(n->sxmn + n->sxmx)/2;
    cog[1]=(n->symn + n->symx)/2;
    cog[2]=(n->zmin + n->zmax)/2;
  }
  else {
    cog[0]=(n->xmin + n->xmax)/2;
    cog[1]=(n->ymin + n->ymax)/2;
    cog[2]=(n->zmin + n->zmax)/2;
  }
  
  for(i=0; i<9; i++) N[i] = n->rot[i];
  
  /*
    ta = -R_A_transposed*N_tr*cog + N_tr*(delt+cog)
    trans = R_A_tr*x + ta
    
    To calculate ta we can proceed as follows:
    
    t1 = N_tr*cog
    t2 = -R_A_transposed*t1
    t3 = N_tr*(delt+cog)
    ta = t2+t3
    
    */
  
  float t1[3],t2[3],t3[3];
  /* t1 = N_tr*cog */
  for(i=0;i<3;i++){
    t1[i] = 0.;
    for(k=0;k<3;k++){
      int idxN=k+i*3;
      t1[i] += N[idxN]*cog[k];
    }
  }
  
  /* t2 = -R_A_transposed*t1 */
  for(i=0;i<3;i++){
    t2[i] = 0.;
    for(k=0;k<3;k++){
      int idxR=k+i*3;
      t2[i] += -RA[idxR]*t1[k];
    }
  }
  
  /* t3 = N_tr*(delt+cog) */
  for(i=0;i<3;i++){
    t3[i] = 0.;
    for(k=0;k<3;k++){
      int idxN=k+i*3;
      t3[i] += N[idxN]*(delt[k]+cog[k]);
    }
  }
  
  /*  ta = t2+t3 */
  for(i=0;i<3;i++){
    alpar->ta[i] = t2[i] + t3[i];
  }
  
#ifdef USE_ARTE
  // Modify GEDE volumes
  int j;
  float Grot[9],trans[3];
  // GEDE_rot:=NEDE_rot*R_A
  for(i=0;i<3;i++){
    for(k=0;k<3;k++){
      int idxG=i+k*3;
      Grot[idxG]=0;
      for(j=0;j<3;j++){
	int idxN=i+j*3;
	int idxRA=j+k*3;
	Grot[idxG]=Grot[idxG]+(n->rot[idxN])*RA[idxRA];
      }
    }
  }
  
  //   trans=GEDE_rot*N_transposed*(delt+c.o.g.)-c.o.g
  for(i=0;i<3;i++){
    trans[i]=-cog[i];
    for(k=0;k<3;k++){
      for(j=0;j<3;j++){
	int idxG=i+j*3;
	int idxN=k+j*3;
	trans[i]=trans[i]+Grot[idxG]*(n->rot[idxN])*(delt[k]+cog[k]);
      }
    }
  }

  UpdateGede(gi,Grot,trans);

#endif

  return OTRIValid;
}  

/* Prepare both OtrAlignStruct and OtrAlignParStruct for given gi
   (Only internal usage).
    */
int OtrAlignPrepare(int gi,
		    float dx, float dy, float dz,
		    float a, float b, float c) {
  OtrAlignFill(gi, dx,dy,dz, a,b,c);
  OtrAlignParCalc(gi);
  return OTRIValid;
}

/** Update memory structures, arrays ... from database or ascii files.

  Input information is taken from file referenced by
  environment variable OtrAlignInput.

  Example file for run 12333 can be found in
  /afs/desy.de/group/hera-b/OTR/db/align/run12333.dat

  @return 0 if state ok, else error code
  */
  

int   OtrAlignUpdate() {
#ifdef DEBUG
  printf("OtrAlignUpdate called\n");
#endif
  
  if (OtrAlignClose()==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignUpdate",
			"Error during AlignClose","");
  }
#ifdef DEBUG
  printf("OtrAlignUpdate: OtrAlignClose finished\n");
#endif
  
  int ngede = OtrGeoNumberOfVolumes();
  if (ngede<100 || ngede>4000) {
    char buf[100];
    sprintf(buf,"%d",ngede);
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignUpdate",
			"OtrGeoNumberOfVolumes() returns illegal value: ",buf);
  }

  //----- create OtrAlignStruct - either from ASCII file or database
  char* AlignInput=getenv("OtrAlignInput");
  if (AlignInput && AlignInput[0]!='\0') { 

    numberofalign = ngede;

#ifdef DEBUG
  printf("OtrAlignUpdate: Initializing for numberofalign=%d\n",numberofalign);
#endif

    aligns=(OtrAlignStruct*)realloc(aligns,(numberofalign)*sizeof(OtrAlignStruct));
    
    if (aligns==NULL) {
      return OtrIErrorMsg(OTRIInvalid,"OtrAlignUpdate",
			  "Error reallocating space for aligns","");
    }
    for (int gi=0; gi<ngede; gi++) {
      OtrAlignFill(gi,0.,0.,0.,0.,0.,0.);
    }

#ifdef DEBUG
    printf("OtrAlignUpdate: Opening file %s\n",AlignInput);
#endif
    FILE *Data = fopen(AlignInput,"r");
    if (Data==0) {
      return OtrIErrorMsg(OTRIInvalid,"OtrAlignUpdate",
			  "Cannot open file",AlignInput);
    }
    int rc = OtrAlignScan(Data);
    fclose(Data);
    if (rc==OTRIInvalid) {
      return OtrIErrorMsg(OTRIInvalid,"OtrAlignUpdate",
			  "Error during OtrAlignScan of ",AlignInput);
    }
  }
  else { 
    int version;
    if ((version=OtrDbGetTableVersion(AlignTableDef.name))!=-1) {
      printf("OtrAlignUpdate: Calling OtrDbGet v=%d ...\n",version);fflush(stdout);
      aligns = (OtrAlignStruct*) OtrDbGet(&AlignTableDef,aligns,&numberofalign);
      printf("OtrAlignUpdate: OtrDbGet has been called.\n");
      if (aligns==NULL|| numberofalign==0) {
        return OtrIErrorMsg(OTRIInvalid,"OtrAlignUpdate",
			  "Nothing was returned from OtrDbGet for aligns","");
      }
    }
    else {
      printf("OtrAlignUpdate: Version error: %d\n",version);
      fflush(stdout);
      return OtrIInvalid();
    }
  }

  //----- create OtrAlignParStruct
  alignpars=(OtrAlignParStruct*)realloc(alignpars,
					(numberofalign)*sizeof(OtrAlignParStruct));
  if (alignpars==NULL) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignUpdate",
			"Error reallocating space for alignpars","");
  }
  for (int gi=0; gi<ngede; gi++) {
    OtrAlignParCalc(gi);
  }

  return OTRIValid;
}

/** Save memory structures, arrays ... to database or ascii files.

  @return 0 if state ok, else error code
  */

int   OtrAlignWriteToDb() { 
  if (OtrAlignAlgorithm==OTRIInvalid) return OTRIInvalid;
  if (aligns==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignWriteToDb",
			"OtrAlign is not initialized","");
  }
  printf("numberofalign=%d\n",numberofalign);
  printf("sizeof(OtrAlignStruct)=%d\n",sizeof(OtrAlignStruct));
  printf("sizeof(structfmtAlign)=%d\n",sizeof(structfmtAlign));
  printf("sizeof(OtrDbField)=%d\n",sizeof(OtrDbField));
  
  printf("OtrAlignWriteToDb: Calling OtrDbPut...\n");fflush(stdout);
  OtrDbPut(&AlignTableDef,aligns,numberofalign);
  printf("OtrAlignWriteToDb: OtrDbPut has been called.\n");fflush(stdout);
  
  printf("OtrAlignWriteToDb: database written.\n");
  
  return OTRIInvalid;
  }

/** Close OtrAlign package
  (free allocate memory,
    clean up to be in a state that initialization is necessary).

  @return 0 if state ok, else error code
  */

int   OtrAlignClose()     { 
  if (OtrAlignAlgorithm==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignUpdate",
			"OtrAlign is not initialized","");
  }
  numberofalign=0;
  if (aligns) {
    free(aligns);
    aligns=0;
  }
  if (alignpars) {
    free(alignpars);
    alignpars=0;
  }
  return OTRIValid;
}

/** Read input file and fill structure OtrAlignStruct with data.

  @param  FILE *fp   Input file with alignment information
  @return 0 if state ok, else error code

    */
int   OtrAlignScan(FILE *fp) { 
  if (OtrAlignAlgorithm==OTRIInvalid) return OTRIInvalid;

  char rundate[50];
  char stamp[50];
  rundate[0] = '\0';
  stamp[0] = '\0';
  int run=0;
  int indalign = 0;
  int indnede = 0;
  char str[500];
  while (fgets(str,500,fp)) {
    if (strlen(str)==499) {
      return OtrIErrorMsg(OTRIInvalid,"OtrAlignScan",
			  "Line exceed buffer length","");
    }
    //printf("read line:\"%s\"\n",str);
    
    if (str[0] != '*' && strlen(str)>1) {
      char key[30];
      sscanf(str,"%s",key);
      if (strlen(key)==0) continue;
      // printf("key=\"%s\"\n",key);
      
      if (!strcmp(key,"run")) {
        sscanf(str,"%s %d",key,&run);
      }
      else if (!strcmp(key,"rundate")) {
        //sscanf(str,"%s %s",key,rundate);
        strcpy(rundate,str+8);
      }
      else if (!strcmp(key,"stamp")) {
        //sscanf(str,"%s %s",key,stamp);
        strcpy(stamp,str+6);
      }
      else {
	char hwcstart[13];
	float dx,dy,dz,a,b,c; 
	sscanf(str,"%s %f %f %f %f %f %f",hwcstart,&dx,&dy,&dz,&a,&b,&c);
	indalign++;

	int ncompare = 11;
	if (strlen(hwcstart)<11) ncompare = strlen(hwcstart);

	int seen=0;
	int ngede = OtrGeoNumberOfVolumes();
	for (int gi=0; gi<ngede; gi++) {
	  if (!strncmp(hwcstart,OtrGeoGet(gi)->hwc,ncompare)) {
	    indnede++;
	    seen++;
	    char h[13];
	    strncpy(h,OtrGeoGet(gi)->hwc,12);
	    h[12]='\0';
	    //printf("Matched hwc[%d] = %s\n",gi,h);

	    OtrAlignFill(gi, dx,dy,dz, a,b,c);
	  }
	}
	//printf("read align: %-12s %8f %8f %8f %8f %8f %8f %d\n",
	//       hwcstart,dx,dy,dz,a,b,c,seen);
#ifdef DEBUG
	if (seen==0) {
	  printf("hwc %s was not found in NEDE tables\n",hwcstart);
	}
#endif

      }
    }
  }
#ifdef DEBUG
  printf("\n");
  printf("OtrAlignScan: Found  %d lines with alignment data in input file \n",indalign);
  printf("OtrAlignScan: Filled %d Alignment structure elements\n",indnede);
  if (run)               printf("OtrAlignScan: Run was %d\n",run);
  if (strlen(rundate)>2) printf("OtrAlignScan: RunDate was %s\n",rundate);
  if (strlen(stamp)>2)   printf("OtrAlignScan: Stamp was %s\n",stamp);
  printf("\n");
#endif
  
  return OTRIValid;
}


/** Print memory structure to ascii files.

  @param  FILE *fp   Output file with alignment information
  @return 0 if state ok, else error code
  */
  
int   OtrAlignPrint(FILE *fp)     { 
  if (OtrAlignAlgorithm==OTRIInvalid) return OTRIInvalid;
  if (numberofalign==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignPrint",
                        "OtrAlign is not initialized","");
  }
  
  for(int gi=0; gi<numberofalign; gi++) {
    OtrAlignStruct* al = &aligns[gi];
    OtrAlignParStruct* alpar = &alignpars[gi];
    char h[13];
    strncpy(h, al->hwc,12);
    h[12]='\0';
    int i;
    for (i=0; i<3; i++) if (fabs(al->delt[i]) > 1.e-7) goto print;
    for (i=0; i<3; i++) if (fabs(al->ang[i]) > 1.e-6) goto print;
    for (i=0; i<3; i++) if (fabs(alpar->ta[i]) > 1.e-6) goto print;
    for (i=0; i<9; i++) {
      if (i==0||i==4||i==8) {
	if (fabs(1.-alpar->ra[i]) > 1.e-7) goto print;
      }
      else {
	if (fabs(alpar->ra[i] > 1.e-7)) goto print;
      }
    }
    continue;
    
  print:
    fprintf(fp,"%s ",h);
    for (i=0; i<3; i++) fprintf(fp,"%f ",al->delt[i]);
    for (i=0; i<3; i++) fprintf(fp,"%f ",al->ang[i]);
    
    for (i=0; i<3; i++) fprintf(fp,"%f ",al->eang[i]);
    for (i=0; i<3; i++) fprintf(fp,"%f ",al->edelt[i]);
    for (i=0; i<3; i++) fprintf(fp,"%f ",al->sloff[i]);
    for (i=0; i<3; i++) fprintf(fp,"%f ",al->esloff[i]);
    fprintf(fp,"%f ",al->redres);
    for (i=0; i<9; i++) fprintf(fp,"%f ",alpar->ra[i]);
    for (i=0; i<3; i++) fprintf(fp,"%f ",alpar->ta[i]);
    fprintf(fp,"\n");
  }
  return OTRIValid;
}


/** Test bench to be sucessfully executed before any release of a new version.

  @return 0 if state ok, else error code
  */

int   OtrAlignTestMe()    {
  printf("OtrAlignTestMe called.\n");


#ifdef DEBUG
  FILE *fp=fopen("/dev/tty","w");
  fprintf(fp,"Calling OtrDbPrint...\n");
  OtrDbPrint(fp,&AlignTableDef);
  fclose(fp);
#endif

  printf("OtrAlignTestMe: OtrGeoInitialize() was successful\n");
  
  int ngede = OtrGeoNumberOfVolumes();

#ifdef DEBUG
  printf ("OtrAlignTestMe: ngede=%d\n",ngede);
#endif

  putenv("OtrAlignInput=/afs/desy.de/group/hera-b/OTR/db/align/run12333.dat");
  if (OtrAlignInitialize()==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignTestMe",
			"Error during initialization","");
    }
  
  if (numberofalign==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignTestMe",
                        "OtrAlign is not initialized","");
    }
  
  putenv("OtrAlignOutput=align.test");
  char* AlignOutput=getenv("OtrAlignOutput");
  if (AlignOutput) { 
      printf("Saving alignment structure to output file %s\n",AlignOutput);
      FILE *fp = fopen(AlignOutput,"w");
      if (OtrAlignPrint(fp)==OTRIInvalid) {
        return OtrIErrorMsg(OTRIInvalid,"OtrAlignTestMe",
			  "Error during printing","");
      }
    fclose(fp);
    }
    
  printf("Alignment structure saved to output file %s\n",AlignOutput);

  putenv("OtrAlignInput=");
  OtrDbUserDefine("/OTR_TEST","OTR_Align",OtrAlignUpdate,
                  0x7fffffff,0x7fffffff,2,1);
  OtrAlignWriteToDb();
  OtrAlignClose();
  
  printf("Init from Db\n");
  OtrAlignInitialize();
  {
    printf("Saving alignment structure to output file align.db\n");
    FILE *fp = fopen("align.db","w");
    OtrAlignPrint(fp);
    fclose(fp);
    }
  
  int ind=0;
  int gi;
  char hwc[13];
  float x,y,z,n[3];
  for (gi=0; gi<ngede; gi++) {
    strncpy(hwc,OtrGeoGet(gi)->hwc,12);
    hwc[12]='\0';
    ind++;
    
    x = 0.3;
    y = 0.8;
    z = 0.4;
    n[0] = x;
    n[1] = y;
    n[2] = z;
    OtrAlignRotate(gi,n);
    OtrAlignRotateBack(gi,n);
    if (fabs(n[0]-x)>1.e-4 || fabs(n[1]-y)>1.e-4 || fabs(n[2]-z)>1.e-4 ) {
      printf("OtrAlignTestMe: Rotation Back did not give initial vector.\n n=%f %f %f\n",
	     n[0],n[1],n[2]);
    }
    
    
    x = 0.3;
    y = 0.8;
    z = 0.4;
    n[0] = x;
    n[1] = y;
    n[2] = z;
    OtrAlignShift(gi,n);
    OtrAlignShiftBack(gi,n);
    if (fabs(n[0]-x)>1.e-4 || fabs(n[1]-y)>1.e-4 || fabs(n[2]-z)>1.e-4 ) {
      printf("OtrAlignTestMe: Shift Back did not give initial vector.\n");
      printf("hwc=%s x=%f %f %f\n",hwc,n[0],n[1],n[2]);
      
    }
  }
  printf("Back functions checked.\n");

  gi = OtrIGeoIndexFromModule(9,2,6,1); // -p1-2-06-01
#ifdef DEBUG
  printf("gi=%d corresponds to Volume -p1-2-06-01\n",gi);
#endif
  if (gi<=0) {
    gi = 100;
#ifdef DEBUG
    printf("Using gi=%d corresponding to GEDE %12.12s\n",
	   gi,OtrGeoGet(gi)->hwc);
#endif
  }

#ifdef USE_ARTE
  ArtePointer<NEDE> nede = OtrGeoGet(gi);
#else
  NEDE* nede = (NEDE*)OtrGeoGet(gi);
#endif
  strncpy(hwc,nede->hwc,12);
  hwc[12]='\0';
  float cog[3];
  cog[0]=(nede->sxmn + nede->sxmx)/2;
  cog[1]=(nede->symn + nede->symx)/2;
  cog[2]=(nede->zmin + nede->zmax)/2;
  float r1 = nede->rot[1];
#ifdef DEBUG
  printf("\nTest of Alignment for gi=%d hwc=%s: x0=%f %f %f, rot[1]=%f\n",
	 gi,hwc,cog[0],cog[1],cog[2], r1);
#endif
  
#ifdef DEBUG
  printf("Translate NEDE by dx = 1 cm\n");
#endif
  OtrAlignPrepare(gi, -1.,0.,0., 0.,0.,0.);
  x = 0.3;
  y = 0.8;
  z = 0.4;
  n[0] = x;
  n[1] = y;
  n[2] = z;
  OtrAlignRotate(gi,n);
  if (fabs(n[0]-x) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[0]=%f not equal to %f\n", n[0], x);
  }
  if (fabs(n[1]-y) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[1]=%f not equal to %f\n", n[1], y);
  }
  if (fabs(n[2]-z) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[2]=%f not equal to %f\n", n[2], z);
  }
#ifdef DEBUG
  printf("After OtrAlignRotate: x=%f %f %f\n",n[0],n[1],n[2]);
#endif

  n[0] = x;
  n[1] = y;
  n[2] = z;
  OtrAlignShift(gi,n);
  if (fabs(n[0]-x-sqrt(1.-r1*r1)) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[0]=%f not equal to %f\n", n[0], x+sqrt(1.-r1*r1));
  }
  if (fabs(n[1]-y+r1) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[1]=%f not equal to %f\n", n[1], y-r1);
  }
  if (fabs(n[2]-z) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[2]=%f not equal to %f\n", n[2], z);
  }
#ifdef DEBUG
  printf("After OtrAlignShift : x=%f %f %f\n\n",n[0],n[1],n[2]);
#endif

#ifdef DEBUG
  printf("Translate NEDE by dz = 1 cm\n");
#endif
  OtrAlignPrepare(gi, 0.,0.,-1., 0.,0.,0.);
  x = 0.3;
  y = 0.8;
  z = 0.4;
  n[0] = x;
  n[1] = y;
  n[2] = z;
  OtrAlignRotate(gi,n);
  if (fabs(n[0]-x) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[0]=%f not equal to %f\n", n[0], x);
  }
  if (fabs(n[1]-y) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[1]=%f not equal to %f\n", n[1], y);
  }
  if (fabs(n[2]-z) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[2]=%f not equal to %f\n", n[2], z);
  }
#ifdef DEBUG
  printf("After OtrAlignRotate: x=%f %f %f\n",n[0],n[1],n[2]);
#endif

  n[0] = x;
  n[1] = y;
  n[2] = z;
  OtrAlignShift(gi,n);
  if (fabs(n[0]-x) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[0]=%f not equal to %f\n", n[0], x);
  }
  if (fabs(n[1]-y) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[1]=%f not equal to %f\n", n[1], y);
  }
  if (fabs(n[2]-z-1.) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[2]=%f not equal to %f\n", n[2], z+1.);
  }
#ifdef DEBUG
  printf("After OtrAlignShift : x=%f %f %f\n\n",n[0],n[1],n[2]);
#endif

  float a=0.088;
#ifdef DEBUG
  printf("Rotate NEDE around Z axis by c = %f rad\n",a);
#endif
  OtrAlignPrepare(gi, 0.,0.,0., 0.,0.,a);
  x = 0.3;
  y = 0.8;
  z = 0.4;
  n[0] = x;
  n[1] = y;
  n[2] = z;
  OtrAlignRotate(gi,n);
  if (fabs(n[0]-x*cos(a)+y*sin(a)) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[0]=%f not equal to %f\n", n[0], x*cos(a)-y*sin(a));
  }
  if (fabs(n[1]-x*sin(a)-y*cos(a)) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[1]=%f not equal to %f\n", n[1], x*sin(a)+y*cos(a));
  }
  if (fabs(n[2]-z) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[2]=%f not equal to %f\n", n[2], z);
  }
#ifdef DEBUG
  printf("After OtrAlignRotate: x=%f %f %f\n",n[0],n[1],n[2]);
#endif

  n[0] = x;
  n[1] = y;
  n[2] = z;
  OtrAlignShift(gi,n);
#ifdef DEBUG
  printf("After OtrAlignShift : x=%f %f %f\n\n",n[0],n[1],n[2]);
#endif


#ifdef DEBUG
  printf("Rotate NEDE around X axis by a = %f rad\n",a);
#endif
  OtrAlignPrepare(gi, 0.,0.,0., a,0.,0.);
  x = 0.3;
  y = 0.8;
  z = 0.4;
  n[0] = x;
  n[1] = y;
  n[2] = z;
  OtrAlignRotate(gi,n);
  if (fabs(n[0]-x) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[0]=%f not equal to %f\n", n[0], x);
  }
  if (fabs(n[1]-y*cos(a)+z*sin(a)) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[1]=%f not equal to %f\n", n[1], y*cos(a)-z*sin(a));
  }
  if (fabs(n[2]-y*sin(a)-z*cos(a)) > 1.e-5) {
    printf("OtrAlignTestMe: ERROR - n[2]=%f not equal to %f\n", n[2], y*sin(a)+z*cos(a));
  }
#ifdef DEBUG
  printf("After OtrAlignRotate: x=%f %f %f\n",n[0],n[1],n[2]);
#endif

  n[0] = x;
  n[1] = y;
  n[2] = z;
  OtrAlignShift(gi,n);
#ifdef DEBUG
  printf("After OtrAlignShift : x=%f %f %f\n\n",n[0],n[1],n[2]);
#endif

  if (OtrAlignInitialize()==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignTestMe",
			"Error during initialization","");
  }
  
  if (numberofalign==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignTestMe",
                        "OtrAlign is not initialized","");
  }
  
  printf("OtrAlignTestMe: test finished.\n");
  return OTRIValid;
}

/* Test bench 2 - the same but without proper initialization
   - should give an error.

  @return 0 if state ok, else error code
  */

extern "C" int   OtrAlignTestMe2() {
  
  printf("OtrAlignTestMe2 called.\n");
  if (OtrGeoInitialize()==OTRIInvalid) {
    printf("OtrGeoTestMe2: Error during initialization\n");
    return OTRIInvalid;
  }
  printf("OtrAlignTestMe2: OtrGeoInitialize() was successful\n");

  int ngede = OtrGeoNumberOfVolumes();
  printf ("OtrAlignTestMe2: ngede=%d\n",ngede);

  if (OtrAlignInitialize()==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignTestMe2",
			"Error during initialization","");
  }

  if (numberofalign==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignTestMe2",
                        "OtrAlign is not initialized","");
  }

  putenv("OtrAlignOutput=align.out");
  char* AlignOutput=getenv("OtrAlignOutput");
  if (AlignOutput) { 
    printf("Saving alignment structure to output file %s\n",AlignOutput);
    FILE *fp = fopen(AlignOutput,"w");
    if (OtrAlignPrint(fp)==OTRIInvalid) {
      return OtrIErrorMsg(OTRIInvalid,"OtrAlignTestMe2",
			  "Error during printing","");
    }
    fclose(fp);
  }
  printf("Saved to output file.\n");

  int ind=0;
  for (int gi=0; gi<ngede; gi++) {
    char hwc[13];
    strncpy(hwc,OtrGeoGet(gi)->hwc,12);
    hwc[12]='\0';
    ind++;
    float x,y,z,n[3];
    
    x = 0.3;
    y = 0.8;
    z = 0.4;
    n[0] = x;
    n[1] = y;
    n[2] = z;
    OtrAlignRotate(gi,n);
    OtrAlignRotateBack(gi,n);
    if (fabs(n[0]-x)>1.e-4 || fabs(n[1]-y)>1.e-4 || fabs(n[2]-z)>1.e-4 ) {
      printf("OtrAlignTestMe2: Rotation Back did not give initial vector.\n n=%f %f %f\n",
	     n[0],n[1],n[2]);
    }
    
    
    x = 0.3;
    y = 0.8;
    z = 0.4;
    n[0] = x;
    n[1] = y;
    n[2] = z;
    OtrAlignShift(gi,n);
    OtrAlignShiftBack(gi,n);
    if (fabs(n[0]-x)>1.e-4 || fabs(n[1]-y)>1.e-4 || fabs(n[2]-z)>1.e-4 ) {
      printf("OtrAlignTestMe2: Shift Back did not give initial vector.\n");
      printf("hwc=%s x=%f %f %f\n",hwc,n[0],n[1],n[2]);
      
    }
  }
  printf("OtrAlignTestMe2: test finished.\n");
  return OTRIValid;
}

/**
Transform direction vector from non-aligned to aligned direction.
Coordinates are given in laboratory system.

  @param gi   GeoIndex specification of NEDE
  @param n[3] 3-vector to be transformed (both input and output)
  @return 0 if state ok, else error code
*/

int OtrAlignRotate(int gi,float n[3]) {
  if (numberofalign==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignRotate",
                        "OtrAlign is not initialized","");
  }
  
  OtrAlignParStruct* alpar = &alignpars[gi];

  /* x_ACS_Aligned := R_A * x_ACS_Nonaligned */
  int i,k;
  float na[3];
  for(i=0;i<3;i++){
    na[i]=0;
    for(k=0;k<3;k++){
      int idxR=i+k*3;
      na[i] = na[i] + (alpar->ra[idxR])*n[k];
    }
  }
  n[0] = na[0];
  n[1] = na[1];
  n[2] = na[2];
  return OTRIValid;
};

/**
Transform direction vector from aligned to non-aligned direction.
Coordinates are given in laboratory system.

  @param gi   GeoIndex specification of NEDE
  @param n[3] 3-vector to be transformed (both input and output)
  @return 0 if state ok, else error code
*/

int OtrAlignRotateBack(int gi,float n[3]) {
  if (numberofalign==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignRotateBack",
                        "OtrAlign is not initialized","");
  }
  
  OtrAlignParStruct* alpar = &alignpars[gi];

  /* x_ACS_nonaligned := R_A_transposed * x_ACS_Aligned */
  int i,k;
  float na[3];
  for(i=0;i<3;i++){
    na[i]=0;
    for(k=0;k<3;k++){
      int idxR=k+i*3;
      na[i] = na[i] + (alpar->ra[idxR])*n[k];
    }
  }
  n[0] = na[0];
  n[1] = na[1];
  n[2] = na[2];
  return OTRIValid;
};

/**
Transform X,Y,Z-vector from non-aligned to aligned.
Coordinates are given in laboratory system.

  @param gi   GeoIndex specification of NEDE
  @param x[3] 3-vector to be transformed (both input and output)
  @return 0 if state ok, else error code
*/

int OtrAlignShift(int gi,float x[3]) {
  if (numberofalign==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignShift",
                        "OtrAlign is not initialized","");
  }
  
  OtrAlignParStruct* alpar = &alignpars[gi];

  /* x_ACS_Aligned := R_A * x_ACS_Nonaligned + ta */
  int i,k;
  float xa[3];
  for(i=0;i<3;i++){
    xa[i]=alpar->ta[i];
    for(k=0;k<3;k++){
      int idxR=i+k*3;
      xa[i] = xa[i] + (alpar->ra[idxR])*x[k];
    }
  }
  x[0] = xa[0];
  x[1] = xa[1];
  x[2] = xa[2];

  return OTRIValid;
};

/**
Transform X,Y,Z-vector from aligned to non-aligned.
Coordinates are given in laboratory system.

  @param gi   GeoIndex specification of NEDE
  @param x[3] 3-vector to be transformed (both input and output)
  @return 0 if state ok, else error code
*/

int OtrAlignShiftBack(int gi,float x[3]) {
  if (numberofalign==0) {
    return OtrIErrorMsg(OTRIInvalid,"OtrAlignShiftBack",
                        "OtrAlign is not initialized","");
  }
  
  OtrAlignParStruct* alpar = &alignpars[gi];

  /* x_ACS_Nonaligned := R_A_transposed * (x_ACS_Aligned - ta) */
  int i,k;
  float xa[3];
  for(i=0;i<3;i++){
    xa[i]=0;
    for(k=0;k<3;k++){
      int idxR=k+i*3;
      xa[i] = xa[i] + (alpar->ra[idxR]) * (x[k] - alpar->ta[i]);
    }
  }
  x[0] = xa[0];
  x[1] = xa[1];
  x[2] = xa[2];

  return OTRIValid;
};

/* ----------------------------------------------------------------------*/
/* ---------------------- KUIP routines ---------------------------------*/

#ifdef USE_ARTE
extern "C" {
  extern char* ku_gets();
  extern int ku_geti();
  
/** KUIP entry for OtrAlignSetAlgorithm(int i)
 */
  void OtrAlignSetAlgorithm_kuip() {
    int i = ku_geti();
    printf("Setting OtrAlignAlgorithm to %d.\n",i);
    OtrAlignSwitch(i);
  }
}

#endif

/*@}*/
