#ifndef __OtrGeo__H__
#define __OtrGeo__H__

#include <stdio.h>

#ifdef USE_ARTE
#  include "arte/arte.hh" 
#else
#include "auto/NEDE.h"
#include "auto/GEDE.h"
#  undef JNEDE 
#  undef NNEDE
#  undef NCNEDE
#  undef NEDE
#  undef NEDE_onesl
#  undef NEDE_ogmat

#  undef JGEDE
#  undef NGEDE
#  undef NCGEDE
#  undef GEDE
#  undef GEDE_ogesl
#  undef GEDE_ogmat
#  undef GEDE_ootmo

#endif

#ifdef __cplusplus
extern "C" {
#endif

/** @name OtrGeo 
  Interface to the NEDE structure
  
  */
/*@{*/
  
/* ----------------------------------------------------------------------*/
/**
  Switch between different solutions of the task of the package.

 @param  i Algorithm (currently available only 0)
  */

void  OtrGeoSetAlgorithm(int i);


/**
  Tasks:
    Call OTrDbDefine to declare data base tables

  */

void   OtrGeoDefine();

/**
  Tasks:
    Memory allocation,
    prepare for reinitialization,
    after this call reinitalization.

  All function of the package (beside OtrGeoSetAlgorithm) have to check,
  if the initialization was done properly.
  */

int   OtrGeoInitialize();

/** Prepare lookup table for plate number on which the module resides.
 */
void   OtrGeoUpdatePlate();

/** Update memory structures, arrays ... from database or ascii files.
 */
int   OtrGeoUpdate();

/** Provide output memory structures, arrays ... to database or ascii files. */
int   OtrGeoWriteToDb();

/** Tasks: free allocate memory,
    clean up to be in a state that initialization is necessary.

 @return OtrIValid() in case of successful execution
 */

int   OtrGeoClose();

/** Print NEDE structure to an ASCII file.

 @param  *fp Opened file
 */
int   OtrGeoPrint(FILE *fp);

/** Read NEDE structure from an ASCII file.

 @param  *fp Opened file
 */
int   OtrGeoScan(FILE *fp);

/** Gives plate number on which the module resides. Useful for alignment purposes.

  @param  sl Superlayer index (1..30)
  @param  lay Stereo layer    (1..6)
  @param  mod Module No. inside stereo layer (1..39)

  @return No. of plate
  */
int OtrGeoModulePlate(int sl,int lay,int mod);

/** Test bench to be sucessfully executed before any release of a new version.

 @return 0 in case of successful execution
 */
int   OtrGeoTestMe();

/** Test bench for OtrGeoModulePlate.
 */
int   OtrGeoPlateTestMe();

int   OtrGeoNumberOfVolumes();

/** OTR geometry identifier (=hwc in NEDE).
   
 @param  gi GeometryIndex  (0.. (OtrGeoNumberOfVolumes()-1))
 @param hwc  geometry identifier at return
  */
void OtrGeoId(int gi,char hwc[13]);

/** OTR superlayer of volume.
   
 @param  gi GeometryIndex  (0.. (OtrGeoNumberOfVolumes()-1))
 @return SuperLayer
  */
int OtrGeoSuperLayer(int gi);

/** OTR layer of volume.
   
 @param  gi GeometryIndex  (0.. (OtrGeoNumberOfVolumes()-1))
 @return Layer
  */
int OtrGeoLayer(int gi);

/** OTR sector of volume.
   
 @param  gi GeometryIndex  (0.. (OtrGeoNumberOfVolumes()-1))
 @return sector
  */
int OtrGeoSector(int gi);

/** OTR module of volume.
   
 @param  gi GeometryIndex  (0.. (OtrGeoNumberOfVolumes()-1))
 @return Module
  */
int OtrGeoModule(int gi);


/** OTR ARTE-component number.
   
 @param  gi GeometryIndex  (0.. (OtrGeoNumberOfVolumes()-1))
 @return component
  */
int OtrGeoCmp();

/** Get borders of given volume in Laboratory system.

 @param  gi GeometryIndex  (0.. (OtrGeoNumberOfVolumes()-1))
 @param  c1[3] vector of minimum values
 @param  c2[3] vector of maximum values
 */
void OtrGeoVolume(int gi,float c1[3], float c2[3]);

/** Structure of OTrGeoMatch data */
struct OtrGeoMatchStructDef {
  /// Superlayer Id (1-30)
  int sl;
  /// Stereolayer (1-6)
  int lay;
  /// Flag (currently unused)
  int flag;
  /// Limit of required overlap on Y coordinate (recommended value: 7 cm)
  float ylimit;
  /// GeoIndex of module No.1 (0-2331)
  int gi1;
  /// GeoIndex of module No.2 (1-2332)
  int gi2;
  /// Channel index of channel in module No.1
  int ci1;
  /// Channel index of channel in module No.2
  int ci2;
  /// Module No.1
  int mod1;
  /// Module No.2
  int mod2;
  /// Monolayer of channel in module No.1 (0-3)
  int mono1;
  /// Monolayer of channel in module No.2 (0-3)
  int mono2;
  /// Wire number in monolayer of channel in module No.1 (0-31)
  int wire1;
  /// Wire number in monolayer of channel in module No.2 (0-31)
  int wire2;
  /// X-Difference in cm (with sign)
  float dx;
  /// Z-Difference in cm (with sign)
  float dz;
  /// Maximum allowed distance on X (by default: half of pitch)
  float dxmax;
  /// GeoIndex of first module of this stereolayer
  int gifirst;
  /// GeoIndex of last module of this stereolayer
  int gilast;
  /// X coordinate of matched wire in module No.1
  float x1;
  /// X coordinate of matched wire in module No.2
  float x2;
};

/** Structure of OTrGeoMatch data - typydef */
typedef struct OtrGeoMatchStructDef OtrGeoMatchStruct;

/** Initialize the usage of OtrGeoMatch.

 @param  sl                 Superlayer id  (0..30)
 @param  lay                Stereolayer    (1..6)
 @param  flag               (Currently not used)
 @param  ylimit             Minimum overlap of wires in y to consider them matching ones
 @param  OtrGeoMatchStruct*
 @return 0
 */
int OtrGeoMatchInitialize(int sl,int lay,int flag,float ylimit,OtrGeoMatchStruct *s);

/** Find next match of channels in the given layer.

  Channels are defined to match if they have approximately the same X coordinate.

  Please proceed as follows:
<pre>
  // declare structure to store the results of search
  OtrGeoMatchStruct m;

  // Initialize the search with the given superlayer, stereolayer,
  //   flag, Y limit, and the structure
  OtrGeoMatchInitialize(sl, lay, 0, 7., &m);  

  // do the loop
  while (OtrGeoMatch(&m)) {

    // results of search can be retrived using the structure data
    float x1 = OtrGeoGet(m.gi1)->xfw[m.mono1]+m.wire1*OtrGeoGet(m.gi1)->pit;
    float x2 = OtrGeoGet(m.gi2)->xfw[m.mono2]+m.wire2*OtrGeoGet(m.gi2)->pit;

    printf("%3d: gi1 %4d \"%12.12s\", mono1 %d, w1 %2d, x=%7.2f <==> "
                "gi2 %4d \"%12.12s\", mono2 %d, w2 %2d, x=%7.2f\n",
           i, m.gi1, OtrGeoGet(m.gi1)->hwc, m.mono1, m.wire1, m.x1,
              m.gi2, OtrGeoGet(m.gi2)->hwc, m.mono2, m.wire2, m.x2);
  }
</pre>


 @param  OtrGeoMatchStruct *s
 @return 1 if match found, 0 if no more matches exist
 */
int OtrGeoMatch(OtrGeoMatchStruct *s);

/** Test bench for OtrGeoMatch.

 @return 0 in case of successful execution
 */
int OtrGeoMatchTestMe();

/*@}*/

#ifndef USE_ARTE
const NEDE* OtrGeoGet(int gi);
#endif

#ifdef __cplusplus
};
#endif


#ifdef USE_ARTE
ArtePointer<NEDE> OtrGeoGet(int gi);
#endif

#endif
