#ifndef __OTRI_H__
#define __OTRI_H__

/** @name OtrI 

    The OtrI interface is made to allow easy access to all information of a given channel. 
    The biggest advantage of a
    interface based on function calls is, that changes in the representation 
    of the data in memory (i.e. database structure) will not be visible 
    for the outside users. 
    The general idea is the following. During run time
    each channel is defined uniquely through a channel index. These index
    is choosen to allow fast access on the decoding, calibration and geometry
    information. The value of the channel index is allowed to change maybe
    due to different code versions or changes in the hardware. Since in the OTR
    a channel might be addressed in several different ways,  a few functions are 
    provided to calculate the channelindex as function of the definition
    OtrIFromFed(fedid,tdcchannel) or OtrIFromModule(superlayer,layer,sector,module,monolayer,wirenumber) ...
    In addition to the channel index a ChannelId is defined. It is meant to
    define a channel uniquely indepent of code versions and hardware changes.
    This ChannelId can be stored in DST-files or in the data bases.
    */
/*@{*/
/*@}*/

#include <stdio.h>  
#include <stdlib.h>  
#include <string.h>  
#include <math.h>  
#include "Otr/OtrDecode.h"
#include "Otr/OtrGeo.h"
#include "Otr/OtrAlign.h"
#include "Otr/OtrDq.h"
#include "Otr/OtrRt.h"
#include "Otr/OtrT0.h"

#ifdef USE_ARTE
#include "arte/arte.hh"
#endif

typedef OtrDecodeStruct OtrDecode;


/* Convention for macro:  only package name in big letter  ? */  

#ifdef __cplusplus
extern "C" {
#endif
int OtrIGetTables(int runno,int time,const OtrDbTableStruct **t,int *n);

extern int otrierrorflag;

#define OTRIValid 0
#define OTRIInvalid -1
#define OTRINotConfigured -2
#define OTRIOutOfRange -3


extern int otridebugtostd;

enum OtrIDebugFlag {
    OTRIDebug        =0x0001,
    OTRIInfo         =0x0002,
    OTRIWarning      =0x0004,
    OTRIError        =0x0008,
    OTRISevere       =0x0010,
    OTRIFatal        =0x0020,    
    OTRIAll          =0x003f,    
    OTRIDebugToStdout=0x01000000,
    OTRIDebugToStderr=0x02000000,
    OTRIDebugToStd   =0x03000000,
    OTRIDebugToGrr   =0x04000000,
    OTRIDebugToAll   =0x07000000,
    OTRIDebugDefault =OTRIDebugToStdout|OTRIAll,
    OTRIDebugOnline  =OTRIDebugToGrr|(OTRIAll&~OTRIDebug),
    OTRIDebugOffline =OTRIDebugToStdout|(OTRIAll&~OTRIDebug)
    };

extern void OtrIDebugSwitch( int i );

#define OTRIUseEnvironment     0x1
#define OTRIUseDb              0x2
#define OTRIUseDefaults        0x4

extern void OtrISwitch(int i);

extern int OtrIInitialize();
extern int OtrITestMe();


/* be aware, that compiler do not check the consistency of format and arg */

extern int otri_errormsg(const char *fcnname,int severity, int errorflag,char *msg);
extern int otri_printf(char *format, ...);
extern int otri_fprintf(FILE* fp,char *format, ...);

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------*/
/* Error handling                                                         */

/** Access Error State.
 @return 0 if state ok, else error code
  */
inline int OtrIError() { return otrierrorflag; }
inline int OtrIValid() { return (otrierrorflag=OTRIValid); }
inline int OtrIInvalid() { return (otrierrorflag=OTRIInvalid); }

/* map all printf and fprintf to otri_errmsg handling */

#ifdef OTRIDebugOut
#define printf   if (otridebugtostd) otri_printf
#define fprintf  otri_fprintf
#endif

inline int OtrIErrorMsg(int type,const char *function,char *msg,char *msgarg) {
  char *s=(char*) malloc(strlen(msg)+strlen(msgarg)+1);
  sprintf(s,"%s %s",msg,msgarg);
  otri_errormsg(function,OTRIFatal,type,s);
  free(s);
  return (otrierrorflag=type);
  }


/* -----------------------------------------------------------------------*/

/*
inline unsigned char OtrISuperLayer(int ci);
inline unsigned char OtrILayer(int ci);
inline unsigned char OtrISector(int ci);
inline unsigned char OtrIModule(int ci);
inline unsigned char OtrIMonoLayer(int ci);
inline unsigned char OtrIWire(int ci);
inline unsigned char OtrIAsd(int ci);
inline int           OtrITdcNumberOfChannel();
*/
inline int           OtrIGeoIndexFromModule(int sl,int l,int mod,int sec);

/** @name OtrI-ChannelId */
/*@{*/
/** OTR channel identifier.

   bit  0-11:  module+100*asd identifier.
   bit 12-17:  wire number in monolayer.
   bit 18-19:  monolayer.
   bit 20-23:  sector.
   bit 24-26:  layer.
   bit 27-31:  super layer.
   
   For coding and decoding of this identifier use always the following official
   coding and decoding function. The sector id is the one used in the specification
   of the ASD8 location. Since there are connections from one module to another,
   the module the channel is belonging to, might be in a different sector.
    
 @param  superlayer  superlayer id 
 @param  layer       layer id 
 @param  sector      sector id 
 @param  monolayer   monolayer id 
 @param  wire        wire id 
 @return ChannelId
  */

inline int OtrIChannelId(int superlayer,int layer,int sector,int module,int monolayer,
                          int wire,int asdid) {
  return (superlayer<<27)|(layer<<24)|(sector<<20)|
         (monolayer<<18)|(wire<<13)|(module<<7)|asdid;
  } 
  
/** OTR channel identifier.
   
 @param  ci ChannelIndex 
 @return ChannelId
  */

inline int OtrIdFromIndex(int ci) {
/*
  return OtrIChannelId(OtrISuperLayer(ci),OtrILayer(ci),OtrISector(ci),
                   OtrIModule(ci),OtrIMonoLayer(ci),OtrIWire(ci),OtrIAsd(ci));
	*/	   
  int chid=OtrDecodeGet()->ASD8s[(ci&0xffff)].wire[(ci>>16)].ChID;
  OtrIValid();
  return (chid!=0) ? chid : OtrIInvalid();
  } 


/* unpacking:  
  return (chid>>27)&0x1F;    super layer 
  return (chid>>24)&0x07;    layer       
  return (chid>>20)&0x0F;    sector      
  return (chid>>18)&0x03;    mono layer  
  return (chid>>13)&0x1F;    wire        
  return (chid>> 7)&0x3F;  module      
  return (chid&0x7F)   ;  asdid  
  */

/** decode OTR superlayer number from ChannelId.   

 @param  chid ChannelId 
 @return Superlayer.
  */
inline int OtrISuperLayerFromId(int chid) {
  return (chid>>27)&0x1F; 
  }
  
/** decode OTR layer number from ChannelId.   

 @param  chid ChannelId 
 @return Layer.
  */
inline int OtrILayerFromId(int chid) {
  return (chid>>24)&0x07;    
  }
/** decode OTR sector number from ChannelId. 

  The sector ID is defined like the one used in the description of the ASD location 

 @param  chid ChannelId 
 @return Sector.
  */
inline int OtrISectorFromId(int chid) {
  return (chid>>20)&0x0F;    
  }
/** decode OTR module number from ChannelId.   

 @param  chid ChannelId 
 @return Module.
  */
inline int OtrIMonoLayerFromId(int chid) {
  return (chid>>18)&0x03; 
  }
/** decode OTR monolayer number from ChannelId.   

 @param  chid ChannLelId 
 @return Mono layer number.
  */
inline int OtrIWireFromId(int chid) {
  return (chid>>13)&0x1F;     
  }
/** decode OTR wire number from ChannelId.   

 @param  chid ChannelId 
 @return Wire number.
  */
inline int OtrIModuleFromId(int chid) {
  return (chid>>7)&0x3F;  
  }
/** decode OTR asd identifier from ChannelId.   

 @param  chid ChannelId 
 @return Asd identifier.
  */
inline int OtrIAsdFromId(int chid) {
  return (chid&0x7F);
  }

/*@}*/
/** get geometry index from geometry ID (so called Hardware Code (hwc) in NEDE).   

 @param  ci channel index 
 @return geometry index
  */ 
      
inline int OtrIGeoIndexFromId(char hwc[12]) {
  int i;
  OtrIValid();
  for (i=0;i<OtrGeoNumberOfVolumes();i++) {
    if (!memcmp(hwc,OtrGeoGet(i)->hwc,11)) break;
    }
  if (i==OtrGeoNumberOfVolumes()||OtrGeoGet(i)->hwc[11]=='n') {
    i=OtrIInvalid();
    }
  return i;
  }

/** access geometry identifier (=hwc in NEDE table).  

 @param  superlayer  superlayer id
 @param  layer   layer id
 @param  sector  sector id
 @param  module  module id
 @param  hwc Geometry Identifier (=hwc in NEDE table)
  */     
  
inline void OtrIGeoIdFromModule(int superlayer,int layer,int module,int sector,char hwc[12]) {
  sprintf(hwc,"%3s-%1d-%02d-%02d",
       OtrDecodeGet()->SLname[superlayer],layer,module,sector);
  }


/** access geometry identifier (=hwc in NEDE table).  

 @param  superlayer  superlayer id
 @param  layer   layer id
 @param  sector  sector id
 @param  module  module id 
 @param  hwc Geometry Identifier (=hwc in NEDE table)
  */     
  
inline int OtrIGeoIndexFromModule(int superlayer,int layer,int module,int sector) {
  char hwc[12];
  otrierrorflag=0;
  OtrIGeoIdFromModule(superlayer,layer,module,sector,hwc);
  return OtrIGeoIndexFromId(hwc);
  }

/** @name OtrI-ChannelIndex */
/*@{*/

/* now calc channelindex */

/** derive the channel index from ASD-index and asdchannel.   

 @param  asdindex       (0 .. OtrIAsdNumberOf()-1 )
 @param  asdchannel     (0 .. OtrIAsdNumberOfChannel()-1)
 @return channel index
  */
           
inline int OtrIFromAsd(int asdindex,int asdchannel) {
  OtrIValid();
  if (OtrDecodeGet()->ASD8s[asdindex].wire[asdchannel].ChID==0) OtrIInvalid();
  return (asdchannel<<16)|asdindex;
  }


/** derive the channel index from fedid(=tdcid) and tdc-channel.   

 @param  fedid       (defined as ...)
 @param  fedchannel  (0 .. 127)
 @return channel index on success else OTRINotValid
  */
      
inline int OtrIFromFed(unsigned short fedid,int fedchannel) {
  const OtrDecode* d=OtrDecodeGet();
  int ret=otrierrorflag=OTRIInvalid;
  if (d->NTDCs==0) {
    ret=OtrIErrorMsg(OTRINotConfigured,"OtrIFromFed","No init of TDC-table in OtrDecode","");
    }
  else {
    unsigned short fi=d->FEDs[fedid];
    if(fi!=(unsigned short)0xFFFF&&fedchannel<128) { 
      int asdindex=d->TDCs[fi].ASD[fedchannel>>4];
      int asdchannel=fedchannel&0xf;
      if(asdindex!=0xFFFF&&d->ASD8s[asdindex].wire[asdchannel].ChID!=0) {
        OtrIValid();
        ret=((asdchannel)<<16)|asdindex;
	}
      else {
        ret=(otrierrorflag=OTRIInvalid);
        }
      }
    else {
      ret=OtrIErrorMsg(OTRIInvalid,"OtrIFromFed","FED of channel is not defined","");
      }
    }
  return ret;
  }

/** derive the channel index from tdcid(=fedid) and tdc-channel.   

 @param  tdcid       (defined as ...)
 @param  tdcchannel  (0..127)
 @return channel index on success else OTRINotValid
  */
      
inline int OtrIFromTdc(unsigned short tdcid,int tdcchannel) {
  return OtrIFromFed(tdcid,tdcchannel);
  }


/** derive the channel index from superlayer,layer,sector,asd and asdchannel.   

 @param  superlayer       (defined as ...)
 @param  layer            (defined as ...)
 @param  sector           (defined as ...)
 @param  asd              (defined as ...)
 @param  asdchannel  (0..15)
 @return channel index on success else OTRINotValid
  */
           
inline int OtrIFromAsdId(int superlayer,int layer,int sector,int asd,int asdchannel) {
  const OtrDecode* d=OtrDecodeGet();
  int ret=OtrIInvalid();
  if (d->SLs[superlayer]!=0) {
    const OtrSuperLayerStruct* s=&d->Slayers[superlayer];
    if (s->FirstASD[layer][sector]<=asd) {
      ret=s->Line[s->pos[layer][sector]-s->FirstASD[layer][sector]+asd];
      if (d->ASD8s[ret].wire[asdchannel].ChID!=0) {
        OtrIValid();
        ret=(asdchannel<<16)|ret;
	}
      }
    }
  return ret;
  }
  

/** derive the channel index from superlayer,layer,sector,module, monolayer and wire number.   

 @param  superlayer       (defined as ...)
 @param  layer            (defined as ...)
 @param  sector           (defined as ...)
 @param  module           (defined as ...)
 @param  monolayer        (defined as ...)
 @param  wire             (defined as ...)
 @return channel index on success else OTRINotValid
  */

inline int OtrIFromModule(int superlayer,int layer,int module,int sector,
             int monolayer, int wire) {
  int ret=OtrIInvalid();
  const OtrDecode* d=OtrDecodeGet();
  int gi=OtrIGeoIndexFromModule(superlayer,layer,module,sector);
  int ci;
  
  if (!OtrIError()) {
    ci=d->Geo.CiArray[d->Geo.FirstCellInMonolayer[(gi<<2)|monolayer]+wire];
    if (d->ASD8s[ci&0xffff].wire[(ci>>16)&0xf].ChID!=0) {
      ret=(ci&0xfffff);
      OtrIValid();
      }
    }
  return ret;
  }

/** derive the channel index from ChannelId.   

 @param  ChannelId        (defined as ...)
 @return channel index on success else OTRINotValid
  */

inline int OtrIFromId(int chid) { 
  int ret=OtrIInvalid();
  int asdchannel;
  const OtrASD8Struct* asd;
  int asdindex;
  if (chid!=OTRIInvalid&&chid!=0) {
    asdindex=OtrIFromAsdId(OtrISuperLayerFromId(chid),
                        OtrILayerFromId(chid),
                        OtrISectorFromId(chid),
                        OtrIAsdFromId(chid),
                        0); 
    if (asdindex!=OTRIInvalid) {
      asd=&OtrDecodeGet()->ASD8s[asdindex&0xffff];
      for (asdchannel=0;asdchannel<16;asdchannel++) {
        if (asd->wire[asdchannel].ChID==chid){
          ret=asdindex|(asdchannel<<16);
	  OtrIValid();
          break;
          }
        }
      }
    else ret=OtrIInvalid();
    }
  return ret;
  
  }

/*@}*/
/** @name OtrI-MoveAround */
/*@{*/
  
/** access super layer identifier.   


       minus side     plus side (=minus side+16)

  m1      1               17
  m2      2               18
  m3      3               19
  m4      4               20
  m5      5               21
  m6      6               22
  m8      8               24
  
  p1      9               25
  p2     10               26
  p3     11               27
  p4     12               28

  t1     13               29
  t2     14               30

 @param  ci channel index 
 @return super layer identifier: minus side: 1-14 plus side: 17-30
  */

inline unsigned char  OtrISuperLayer(int ci) {
  return OtrDecodeGet()->ASD8s[(ci&0xffff)].SuperLayer;
  }
    
/** access number of super layers .   

 @param  ci channel index 
 @return number of super layers (=15: valid superlayer index are 1-6,8-14)
  */
inline unsigned char  OtrISuperLayerNumberOf() {
//  return OtrDecodeGet()->ASD8s[(ci&0xffff)].SuperLayer;
  return 32;
  }
    
/** check if super layers id is valid.   

 @param sl  Super Layer Id (for definition look OtrISuperlayer(int ci)) 
 @return 1 if superlayer exist, 0 else.
  */
inline int  OtrISuperLayerExist(int sl) {
  int ssl=sl&0xf;
  return (sl<32)&&(ssl!=0)&&(ssl!=7)&&(ssl!=15);
  }
    
/** access super layer name.   

 @param  ci channel index 
 @return super layer name
  */
inline const char*    OtrISuperLayerName(int ci) {
  return OtrDecodeGet()->SLname[OtrISuperLayer(ci)];
  }
    
/** access super layer name.   

 @param  name  super layer name
 @return super layer id
  */
  
inline int  OtrISuperLayerFromName(const char *name) {
  int ret=OTRIInvalid;
  int i;
  for (i=0;i<32;i++) {
    if (!memcmp(OtrDecodeGet()->SLname[i],name,3)) {
      ret=i;
      break;
      }
    }
  return ret;
  }
    
/** access if channel is on plus or minus side.   

 @param  ci channel index 
 @return return +1 if channel is part of plus chambar, -1 if part of minus chamber 
  */
inline int  OtrISide(int ci) {
  return (OtrDecodeGet()->ASD8s[(ci&0xffff)].SuperLayer<16) ? -1 : +1;
  }

/** access layer identifier.   

 @param  ci channel index 
 @return layer identifier
  */
inline unsigned char  OtrILayer(int ci) {
  return OtrDecodeGet()->ASD8s[(ci&0xffff)].Layer;
  }

/** access number of layers in superlayer.   

 @param  sl super layer index (0 ... (OtrISuperLayerNumberOf()-1)) 
 @return layer identifier
  */
inline int OtrILayerNumberOf(int sl) {
  return OtrDecodeGet()->Slayers[sl].NLayer; 
  }
 
/** check if layer index is valid.   

 @param  sl valid super layer id  
 @param  l  layer id    
 @return 1 if layer id is valid, 0 else.
  */
  
inline int  OtrILayerExist(int sl, int l ) {
  int ret=0;
  if (OtrISuperLayerExist(sl)) {
    if (l!=0&&l<=OtrDecodeGet()->Slayers[sl].NLayer) ret=1;
    }
  return ret;
  }
 
/** access module identifier.   

 @param  ci channel index 
 @return module identifier
  */
inline unsigned char  OtrIModule(int ci) {
  int chid=OtrDecodeGet()->ASD8s[(ci&0xffff)].wire[ci>>16].ChID;
  return OtrIModuleFromId(chid);
  }

/** access number of modules in superlayer.

 @param  sl   valid superlayerindex: 0 ... (OtrISuperLayerNumberOf()-1)
 @param  l    valid layer index:     0 ... (OtrILayerNumberOf(int sl)-1)
 @return number of last module + 1
  */

inline int OtrIModuleNumberOf(int sl,int l) {
  static int nmod[32][6]={
    /* -x side: Below is the number of the last module */
    { 0, 0, 0, 0, 0, 0},  /* not existing */
    { 4, 4, 4, 0, 0, 0},  /* -m1 */
    { 4, 4, 4, 0, 0, 0},  /* -m2 */
    { 5, 5, 5, 0, 0, 0},  /* -m3 */
    { 6, 6, 6, 0, 0, 0},  /* -m4 */
    { 6, 0, 0, 0, 0, 0},  /* -m5 */
    { 7, 7, 7, 0, 0, 0},  /* -m6 */
    { 0, 0, 0, 0, 0, 0},  /* not existing */
    { 8, 8, 8, 0, 0, 0},  /* -m8 */
    {12,12,12,15,15,15},  /* -p1 */
    {12,13,12,12,12,13},  /* -p2: module 1 missing in layer 2&6 */
    {12,13,12,12,12,13},  /* -p3: module 1 missing in layer 2&6 */
    {12,12,12,16,15,15},  /* -p4 */
    {20,18,18, 0, 0, 0},  /* -t1 */
    {21,19,19, 0, 0, 0},  /* -t2 */
    { 0, 0, 0, 0, 0, 0},  /* not existing */

    /* +x side: Below is the number of the last module */
    { 0, 0, 0, 0, 0, 0},  /* not existing */
    { 7, 7, 7, 0, 0, 0},  /* +m1 */
    { 7, 7, 7, 0, 0, 0},  /* +m2 */
    {11,11,11, 0, 0, 0},  /* +m3 */
    {13,13,13, 0, 0, 0},  /* +m4 */
    {12, 0, 0, 0, 0, 0},  /* +m5 */
    {15,15,15, 0, 0, 0},  /* +m6 */
    { 0, 0, 0, 0, 0, 0},  /* not existing */
    {18,18,17, 0, 0, 0},  /* +m8 */
    {21,21,21,27,27,27},  /* +p1 */
    {21,22,21,21,21,22},  /* +p2: last module missing in layer 2&6 */
    {21,22,21,21,21,22},  /* +p3: last module missing in layer 2&6 */
    {21,21,21,29,27,27},  /* +p4 */
    {37,33,33, 0, 0, 0},  /* +t1 */
    {39,35,35, 0, 0, 0},  /* +t2 */
    { 0, 0, 0, 0, 0, 0}   /* not existing */

    };
  return nmod[sl][l-1]+1;
  } 

/** check if module index is valid.   

 @param  sl   0 ... (OtrISuperLayerNumberOf()-1) 
 @param  l    0 ... (OtrILayerNumberOf(int sl)-1) 
 @param  mod  0 ... (OtrIModuleNumberOf(int sl,int l)-1)
 @return 1 if valid, 0 if not existing.
  */

inline int OtrIModuleExist(int sl,int l,int mod) {
  int ret=0;
  if (OtrILayerExist(sl,l)) {
    int n = OtrIModuleNumberOf(sl,l);
    if (mod>0 && mod<n) ret=1;
    if (mod==1 && (sl==10 || sl==11) && (l==2 || l==6)) ret=0;
    if (sl>16 && mod<OtrIModuleNumberOf(sl&0xf,l)) ret=0;
    }
  return ret;
  } 


/** access sector identifier.   

  The sector ID is defined like the one used in the description of the ASD location 

 @param  ci channel index 
 @return sector identifier
  */      
inline unsigned char  OtrISector(int ci) {
  return OtrDecodeGet()->ASD8s[(ci&0xffff)].Sector;
  } 
  
/** access number of sectors in superlayer.   
 @return number of sectors
  */

inline int OtrISectorNumberOf() {
  return 13;
  } 

  
/** check if sector exist.   
 @param  sl   0 ... (OtrISuperLayerNumberOf()-1) 
 @param  l    0 ... (OtrILayerNumberOf(int sl)-1) 
 @param  mod  0 ... (OtrIModuleNumberOf(int sl,int l)-1) 
 @param  sec  0 ... (OtrISectorNumberOf()-1) 
 @return number of sectors
  */

inline int OtrISectorExist(int sl,int l,int mod,int sec) {
  int ret=0;
  if (OtrIModuleExist(sl,l,mod)) {
    OtrIGeoIndexFromModule(sl,l,mod,sec);
    if (!OtrIError()) ret=1;
    }
  return ret;
  } 


/** access mono layer identifier.   

 @param  ci channel index 
 @return mono layer identifier
  */      
inline unsigned char  OtrIMonoLayer(int ci) {
  int chid=OtrDecodeGet()->ASD8s[(ci&0xffff)].wire[ci>>16].ChID;
  return OtrIMonoLayerFromId(chid);
  }
  

/** access wire identifier.   

 @param  ci channel index 
 @return wire identifier
  */     
   
/* how do I know, how many wires are in one mono layer ??? */
/* how do I know, how many monolayer are used ?? gede ?? */
inline unsigned char  OtrIWire(int ci) {
  int chid=OtrDecodeGet()->ASD8s[(ci&0xffff)].wire[ci>>16].ChID;
  return OtrIWireFromId(chid);
  }

/** access asd index.   

 @param  ci channel index 
 @return asd index
  */     
inline unsigned char  OtrIAsdIndex(int ci) {
  return ci&0xffff;
}
  
/** access asd identifier.   

 @param  ci channel index 
 @return asd identifier
  */     
inline unsigned char  OtrIAsd(int ci) {
  return OtrDecodeGet()->ASD8s[(ci&0xffff)].ftb_ASD;
  }
  
/** access asd channel.  

 @param  ci channel index 
 @return asd channel
  */     

inline unsigned char  OtrIAsdChannel(int ci) {
  return (ci>>16);
  }

/** access number of ASD's in the set up.  

 @return number of asd's in the set up
  */     

inline int  OtrIAsdNumberOf() {
  return OtrDecodeGet()->NASDs;
  }

/** access number of ASD's per sector.  

 @param  superlayer Superlayer Index
 @param  layer      Layer Index
 @param  sector     Sector Index
 @return number of ASD's per sector
  */     

inline int  OtrIAsdNumberPerSector(int superlayer,int layer,int sector) {
  return (int)(unsigned int) OtrDecodeGet()->Slayers[superlayer].Nasd[layer][sector];
  }

/** access number of channel on a asd.  

 @return number of channels per asd
  */     

inline int  OtrIAsdNumberOfChannel() {
  return 16;
  }

    
/** access identifier of the fed, where the channel is connected to.  

 @param  ci channel index 
 @return front end driver identifier
  */     
inline unsigned short OtrIFedId(int ci) {
  return OtrDecodeGet()->ASD8s[(ci&0xffff)].TDC_FED;
  }

/** access channel on the fed.  

 @param  ci channel index 
 @return fedchannel (0...127)
  */     
inline int OtrIFedChannel(int ci) {
  return ((OtrDecodeGet()->ASD8s[(ci&0xffff)].TDC_Connector-1)<<4)+(ci>>16);
  }

/** access the number of FED in the set up.  

 @return number of FED's in the set up
  */     
inline int OtrIFedNumberOf() {
  return OtrDecodeGet()->NTDCs;
  }
    
/** access number of channels of a fed.

 @return number of fed channels (=128)
  */     
inline int OtrIFedNumberOfChannel() {
  return 128;
  }
  
/** access fedid by fedindex.  

 @param  fedindex    index of fed     from 0 ...  OtrIFedNumberOf()-1
 @return front end driver identifier
  */     

inline unsigned short OtrIFedIdFromIndex(int fedindex) {
  return OtrDecodeGet()->TDCs[fedindex].FED;
  }

/** access fedindex by fedid.  

 @param  fedid   front end driver identifier
 @return index of fed     from 0 ...  OtrIFedNumberOf()-1 
  */     

inline unsigned short OtrIFedIndexFromId(int fedid) {
  const OtrDecode* d=OtrDecodeGet();
  unsigned short fi=0xffff;
  if (d->NTDCs==0) {
    OtrIErrorMsg(OTRINotConfigured,"OtrIFromFed","No init of TDC-table in OtrDecode","");
    }
  else {
    fi=d->FEDs[fedid];
    OtrIValid();
    }
  if (fi==0xffff) OtrIInvalid();
  return fi;
  }

/** access channel identifier in a loop over all fed & fedchannels.  

 @param  fedindex    index of fed     from 0 ...  OtrIFedNumberOf()
 @param  fedchannel  index of channel from 0 ...  OtrIFedNumberOfChannel()
 @return channel index
  */     
inline int OtrIFromFedIndex(int fedindex,int fedchannel) {
  return OtrIFromFed(OtrIFedIdFromIndex(fedindex),fedchannel);
  }
    
/** access identifier of the tdc=fed, where the channel is connected to.  

 @param  ci channel index 
 @return tdc identifier
  */     
inline unsigned short OtrITdcId(int ci) {
  return OtrIFedId(ci);
  }

/** access the number of FED in the set up.  

 @return number of FED's in the set up
  */     
inline int OtrITdcNumberOf() {
  return OtrIFedNumberOf();
  }
    
/** access number of channels of a tdc.

 @return number of tdc channels (=128)
  */     
inline int OtrITdcNumberOfChannel() {
  return 128;
  }

/** access channel on the TDC.  

 @param  ci channel index 
 @return tdcchannel (0...127)
  */     
inline int OtrITdcChannel(int ci) {
  return OtrIFedChannel(ci);
  }

/** access tdcid(=fedid) by tdcindex (=fedindex).  

 @param  tdcindex    index of fed     from 0 ...  OtrITdcNumberOf()
 @return front end driver identifier
  */     

inline unsigned short OtrITdcIdFromIndex(int tdcindex) {
  return OtrIFedIdFromIndex(tdcindex);
  }

/** access tdcindex (=fedindex)) by tdcid(=fedid.  

 @param   tdcid  front end driver identifier
 @return  index of tdc    from 0 ...  OtrITdcNumberOf()
  */     

inline unsigned short OtrITdcIndexFromId(int tdcid) {
  return OtrIFedIndexFromId(tdcid);
  }

/** access channel identifier in a loop over all fed & fedchannels.  

 @param  fedindex    index of fed     from 0 ...  OtrIFedNumberOf()
 @param  fedchannel  index of channel from 0 ...  OtrIFedNumberOfChannel()
 @return channel index
  */     
inline int OtrIFromTdcIndex(int fedindex,int fedchannel) {
  return OtrIFromFedIndex(fedindex,fedchannel);
  }
       
    
/** access crate, where tdc is plugged into.  

 @param  fedid=tdc identifier 
 @return tdc crate (1..7)
  */     
inline unsigned char  OtrITdcCrateByTdcId(int fedid) {
  return OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[fedid]].crate;
  }
    
/** access slot in crate, where tdc is plugged into.  

 @param  fedid=tdc identifier 
 @return crate slot
  */     
inline unsigned char  OtrITdcSlotByTdcId(int fedid) {
  return OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[fedid]].slot;
  }
    
    
/** access crate, where tdc is plugged into.  

 @param  ci channel index 
 @return tdc crate (1..7)
  */     
inline unsigned char  OtrITdcCrate(int ci) {
  return OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[OtrIFedId(ci)]].crate;
  }
    
/** access slot in crate, where tdc is plugged into.  

 @param  ci channel index 
 @return crate slot
  */     
inline unsigned char  OtrITdcSlot(int ci) {
  return OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[OtrIFedId(ci)]].slot;
  }
    
/** access connector, where asd cable is plugged into.  

 @param  ci channel index 
 @return connector on tdc (0..7)
  */     
inline unsigned char  OtrITdcConnector(int ci) {
  return OtrDecodeGet()->ASD8s[(ci&0xffff)].TDC_Connector;
  }

/** access second level buffer identifier.   
 
 @param  ci channel index 
 @return second level buffer identifier
  */     
inline unsigned char  OtrISlb(int ci) {
  return OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[OtrIFedId(ci)]].SLB;
  }
  
/** access second level buffer board identifier (=SHARC-Board id).   
 
 @param  ci channel index 
 @return second level buffer board identifier
  */     
inline unsigned char  OtrISlbBoard(int ci) {
  return OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[OtrIFedId(ci)]].Sharc_Board;
  }
  
/** access connector on the SLB which is used.  
 
 @param  ci channel index 
 @return connector on SHARC-Board
  */     
    
inline unsigned char  OtrISlbConnector(int ci) {
  return OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[OtrIFedId(ci)]].Sharc_Connector;
  }
    
/** access identifier of the vme, which controls the SLB crate.   
 
 @param  ci channel index 
 @return vme number (e.g. is 66 for hb-vme66)
  */     
inline unsigned char  OtrISlbVme(int ci) {
  return OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[OtrIFedId(ci)]].VME_num;
  }


/*  what is the hv-connector ? The one on the chamber ? */
/** HV-part need to be better implemented.   
 
 @param  ci channel index 
 @return hv connection position at the feedthrough board
  */     
inline unsigned char  OtrIHvConnector(int ci) {
  ci=0;
  return (unsigned char)OTRINotConfigured;
  }
    
inline const char*    OtrIHvGroupName(int ci) {
  int asdchannel=ci>>16;
  const char *c;
  if (asdchannel&0x1) c=OtrDecodeGet()->ASD8s[(ci&0xffff)].HVgroup1;
  else c=OtrDecodeGet()->ASD8s[(ci&0xffff)].HVgroup2;
  return c;
  }
    
inline const char*    OtrIHvCablePanelToDetector(int ci) {
  int asdchannel=ci>>16;
  const char *c;
  if (asdchannel&0x1) c=OtrDecodeGet()->ASD8s[(ci&0xffff)].HVcab1;
  else c=OtrDecodeGet()->ASD8s[(ci&0xffff)].HVcab2;
  return c;
  }
   
/* information about the hv-system is missing:
   power supply channel, grouping, status ...  */ 

  
/*@}*/
/** @name OtrI-Data Access */
/*@{*/

/** check if channel is bad (dead or noisy).   

 @param  ci channel index 
 @return 0 if not dead, else 1
  */     
inline int  OtrIIsBad(int ci) {
  return OtrDqBad(ci);
  }
       
/** check if channel is good (not bad).   

 @param  ci channel index 
 @return 0 if not dead, else 1
  */     
inline int  OtrIIsOk(int ci) {
  return !OtrDqBad(ci);
  }
       
/** check if channel is dead.   

 @param  ci channel index 
 @return 0 if not dead, else 1
  */     
inline int OtrIIsDead(int ci) {
  return OtrDqDead(ci);
  }
       
/** check if channel is noisy.   

 @param  ci channel index 
 @return 0 if not noisy else 1
  */     
inline int  OtrIIsNoisy(int ci) {
  return OtrDqBad(ci)&&!OtrDqDead(ci);
  }
       
    
/** access status of the tdc board.  

 @param  ci channel index 
 @return status of tdc board
  */     
inline unsigned char  OtrITdcStatus(int ci) {
  ci=0;
  return OtrIValid();/*OtrDecodeGet()->TDCs[OtrDecodeGet()->FEDs[OtrIFedId(ci)]].status;*/
  }
        
/** access t0 for the TDC count.  
 
 @param  ci channel index 
 @return t0
  */     
inline float  OtrITdcT0(int ci) {
  return OtrT0Get(ci);
  }
  
/** access error on t0 determination. 
 
 @param  ci channel index 
 @return error of t0
  */     
inline float  OtrITdcT0Sigma(int ci) {
  return OtrT0Sigma(ci);
  }

/** convert tdc-counts into ns value.  
 
 @param  ci channel index 
 @return drifttime in ns
  */     
inline float OtrITdcTime(unsigned char count) {
  return 0.25*(((float)count));
  }
  
/** convert drifttime into distance from wire. 
 
 @param  ci channel index 
 @param  t drifttime in counts
 @return distance from wire in cm
  */     

inline float OtrIRtRadius(int ci,float t)     { 
  return OtrRtRadius(ci,t); 
  }
/** convert drifttime into distance from wire. 
 
 @param  ci channel index 
 @param  t drifttime in counts
 @return error on distance from wire in cm
  */     
inline float OtrIRtRadiusSigma(int ci,float t){ 
  return OtrRtRadiusSigma(ci,t);
  }
/** convert drifttime into distance from wire.  
 
 @param  ci channel index 
 @param  r distance from wire in cm
 @return drifttime in counts
  */     
inline float OtrIRtTime(int ci,float r)       { 
  return OtrRtTime(ci,r);
  }
/** convert drifttime into distance from wire.  
 
 @param  ci channel index 
 @param  r distance from wire in cm
 @return error on drifttime in ns
  */     
inline float OtrIRtTimeError(int ci,float r)  { 
/*  return OtrRtTimeError(float r); */
  ci=0;r=0;
  return OTRIInvalid;
  }
/** access noise probability.
 
 @param  ci channel index 
 @param  count tdc-counts
 @return noise probability 
  */     
inline float OtrIHitNoiseProb(int ci,unsigned char count) { 
  ci=count=0;
  return 0; 
  }
/** access shape of hit efficiency.  
 
 @param  ci channel index 
 @param  r  smallest distance of track from wire 
 @return shape of hit efficiency normalized to 1 
  */     
inline float OtrIHitEfficiencyShape(int ci, float r)       { 
  ci=0;r=0;
  return -1;
  }
  
 /** access hit efficiency.  
 
 @param  ci channel index 
 @return hit efficiency
  */     
inline float OtrIHitEfficiency(int ci)       { 
  ci=0;
  return 1; 
  }
   
/*@}*/
/** @name OtrI-Geometry */
/*@{*/

/** access geometry index.   

 @param  ci channel index 
 @return geometry index
  */     
inline int OtrIGeoIndex(int ci) {
  return OtrDecodeGet()->ASD8s[(ci&0xffff)].wire[ci>>16].NEDE;
  }


  
/** get geometry index from geometry ID (so called Hardware Code (hwc) in NEDE).   

 @param  gi channel index 
 @param  hwc geometry identifier
  */ 
      
inline int OtrIGeoIdFromIndex(int gi,char hwc[12]) {
  int ret;
  if (gi>=0&&gi<OtrGeoNumberOfVolumes()) {
    memcpy(hwc,OtrGeoGet(gi)->hwc,11);
    hwc[11]='\0';
    ret=OtrIValid();
    }
  else {
    hwc[0]='\0';
    ret=otrierrorflag=OTRIInvalid;
    }
  return ret;
  }
  

/** access geometry identifier (=hwc in NEDE table).  

 @param  ci  channel index 
 @param  hwc Geometry Identifier (=hwc in NEDE table)
  */     
  
inline void OtrIGeoId(int ci,char hwc[12]) {
/*
  sprintf(hwc,"%3s-%1d-%02d-%02d",
       &OtrISuperLayerName(ci),OtrILayer(ci),OtrIModule(ci),OtrISector(ci));
       */
  memcpy(hwc,OtrGeoGet(OtrIGeoIndex(ci))->hwc,12);
  hwc[11]='\0';
  }

  

/** access geometry identifier (=hwc in NEDE table).  

 @param  gi  geometry index 
 @return Number of  monolayer in module
  */     
  
inline int OtrIMonoLayerNumberOf(int gi) {
#ifdef USE_ARTE
  ArtePointer<NEDE> nede=OtrGeoGet(gi);
#else
  const NEDE* nede=OtrGeoGet(gi);
#endif
  return (nede->nsl);  
  }

/** access geometry identifier (=hwc in NEDE table).  

 @param  gi  geometry index 
 @return Number of wires in monolayer
  */     
  
inline int OtrIMonoLayerNumberOfWires(int gi,int ml) {
#ifdef USE_ARTE
  ArtePointer<NEDE> nede=OtrGeoGet(gi);
#else
  const NEDE* nede=OtrGeoGet(gi);
#endif
  return (nede->nwi[ml]);  
  }
  
inline int OtrIWiresNumberInMonoLayer(int gi,int ml) {
  return OtrIMonoLayerNumberOfWires(gi,ml);
  }

/** access stereo angle as used/defined in the technical OTR-documentation.   

  The key is to be aware, that the x-component of the direction
  vector of the wire is negative for a +5 degree stereo angle.
   
 @param  ci channel index 
 @return stereo angle
  */     
inline float OtrIWireStereoAngle(int ci) {
  float a=-asin(OtrGeoGet(OtrIGeoIndex(ci))->rot[1]);
  if (fabs(a)<1E-4) a=0;
  return a;
  }

/** access stereo angle as used/defined in the technical OTR-documentation.   

  The FLT uses a more straight forward definition: here the x-component 
  of the direction vector of the wire is positive for a +5 degree stereo 
  angle.
   
 @param  ci channel index 
 @return stereo angle
  */     
inline float OtrIWireStereoAngleFLT(int ci) {
  float a=asin(OtrGeoGet(OtrIGeoIndex(ci))->rot[1]);
  if (fabs(a)<1E-4) a=0;
  return a;
  }

/** access pitch of the wires in the module the channel is belonging to.   


  The pitch is the distance between two wires in one monolayer
  (0.87 cm for 0.5 cm cells, 1.74 cm for 1 cm cells)
   
 @param  ci channel index 
 @return pitch
  */     
  
inline float OtrIWirePitch(int ci) {
  return OtrGeoGet(OtrIGeoIndex(ci))->pit;
  }

/** access the cell diameter.

 @param  ci  channel index 
 @return  cell diameter (0.5 or 1.0 cm)
  */     
  
inline float OtrIGeoCellDiameter(int ci) {
  return (OtrIWirePitch(ci)>1) ? 1.0 : 0.5;
  }

/** access length of the active thin signal wire.   

 @param  ci channel index 
 @return length of avtive part
  */     

inline float OtrIWireLength(int ci) {
#ifdef USE_ARTE
  ArtePointer<NEDE> nede=OtrGeoGet(OtrIGeoIndex(ci));
#else
  const NEDE* nede=OtrGeoGet(OtrIGeoIndex(ci));
#endif

  int ml=OtrIMonoLayer(ci);
  return (nede->y2w[ml]-nede->y1w[ml]);
  }

/* ===================================================================
  Go from NEDE coordinates to Laborary ones:

  x[0] = nede->rot[0]*xrot0 + nede->rot[3]*xrot1 + nede->rot[6]*xrot2;
  x[1] = nede->rot[1]*xrot0 + nede->rot[4]*xrot1 + nede->rot[7]*xrot2;
  x[2] = nede->rot[2]*xrot0 + nede->rot[5]*xrot1 + nede->rot[8]*xrot2;
  
  the origin of the turned coordinate system  for the stereoangle 
  layers is still (x=0,y=0) in the Laboratory frame.

 =================================================================== */
 
/** access all components of the nominal direction of the wire.

 @param  ci channel index 
 @param  n[3] returns all components of the direction of wire (no alignment corrections)
  */     
inline void  OtrIWireDirectionNominal(int ci,float n[3]) {
#ifdef USE_ARTE
  ArtePointer<NEDE> nede=OtrGeoGet(OtrIGeoIndex(ci));
#else
  const NEDE* nede=OtrGeoGet(OtrIGeoIndex(ci));
#endif
  n[0]=nede->rot[1];
  n[1]=nede->rot[0];
  n[2]=0;
  }

/** access all components of the nominal position of the wire in the middle of the active region.

 @param  ci channel index 
 @param  x[3] returns all components of the position of wire (no alignment corrections)
  */       
inline void OtrIWireMiddleNominal(int ci,float x[3]) {
#ifdef USE_ARTE
  ArtePointer<NEDE> nede=OtrGeoGet(OtrIGeoIndex(ci));
#else
  const NEDE* nede=OtrGeoGet(OtrIGeoIndex(ci));
#endif
  int ml=OtrIMonoLayer(ci);
  float xrot0=nede->xfw[ml]+OtrIWire(ci)*nede->pit;
  float xrot1=0.5*(nede->y2w[ml]+nede->y1w[ml]);
  x[0]= nede->rot[0]*xrot0 + nede->rot[1]*xrot1;
  x[1]=-nede->rot[1]*xrot0 + nede->rot[0]*xrot1;
  x[2]=nede->zfw[ml];
  }

  
/** access the direction of the wire in OTR-specific coordinates.

 @param  ci channel index 
 @param  n[3] returns all components of the direction of wire (internal OTR-alignment)
  */     
inline void  OtrIWireDirectionInternal(int ci,float n[3]) {
  int gi=OtrIGeoIndex(ci);
#ifdef USE_ARTE
  ArtePointer<NEDE> nede=OtrGeoGet(gi);
#else
  const NEDE* nede=OtrGeoGet(gi);
#endif
  n[0]=nede->sin;
  n[1]=nede->cos;
  n[2]=0;
  OtrAlignRotate(gi,n);
  }

/** access the position of the wire in the middle of the active region in OTR-specific coordinates.

 @param  ci channel index 
 @param  x[3] returns all components of the position of wire (internal OTR-alignment)
  */       
inline void OtrIWireMiddleInternal(int ci,float x[3]) {
  int gi=OtrIGeoIndex(ci);
#ifdef USE_ARTE
  ArtePointer<NEDE> nede=OtrGeoGet(gi);
#else
  const NEDE* nede=OtrGeoGet(gi);
#endif
  int   ml=OtrIMonoLayer(ci);
  float xrot0=nede->xfw[ml]+OtrIWire(ci)*nede->pit;
  float xrot1=0.5*(nede->y2w[ml]+nede->y1w[ml]);
  x[0]= nede->rot[0]*xrot0 + nede->rot[1]*xrot1;
  x[1]=-nede->rot[1]*xrot0 + nede->rot[0]*xrot1;
  x[2]=nede->zfw[ml];
  OtrAlignShift(gi,x);
  }

/* ===================================================================
  Go from GEDE coordinates to Laboratory ones:

  x[0] = gede->rot[0]*xrot0 + gede->rot[1]*xrot1 + gede->rot[2]*xrot2;
  x[1] = gede->rot[3]*xrot0 + gede->rot[4]*xrot1 + gede->rot[5]*xrot2;
  x[2] = gede->rot[6]*xrot0 + gede->rot[7]*xrot1 + gede->rot[8]*xrot2;

 =================================================================== */
  
  
/** access all components of the nominal direction of the wire.

 @param  ci channel index 
 @param  n[3] returns all components of the direction of wire (global alignment)
  */     
inline void  OtrIWireDirectionGlobal(int ci,float n[3]) {
#ifdef USE_ARTE
  ArtePointer<GEDE> gede(OtrIGeoIndex(ci));
  n[0]=gede->rot[1];
  n[1]=gede->rot[4];
  n[2]=gede->rot[7];
#else
  OtrIWireDirectionInternal(ci,n);
#endif
  }

/** access all components of the nominal position of the wire in the middle of the active region.

 @param  ci channel index 
 @param  x[3] returns all components of the position of wire (global alignment corrections)
  */       
inline void OtrIWireMiddleGlobal(int ci,float x[3]) {
#ifdef NOARTE
  ArtePointer<GEDE> gede=OtrGeoGet(OtrIArteGede(ci));
  int ml=OtrIMonoLayer(ci);
  float xrot0=gede->xfw[ml];
  float xrot1=0.5*(gede->y2w[ml]+gede->y1w[ml]);
  float xrot2=gede->zfw[ml];
  x[0] = gede->rot[0]*xrot0 + gede->rot[1]*xrot1 + gede->rot[2]*xrot2;
  x[1] = gede->rot[3]*xrot0 + gede->rot[4]*xrot1 + gede->rot[5]*xrot2;
  x[2] = gede->rot[6]*xrot0 + gede->rot[7]*xrot1 + gede->rot[8]*xrot2;
#else
  OtrIWireMiddleInternal(ci,x);
#endif
  }

#endif
