/** 
  OtrMon Package:

  Combines packages to fill different levels of monitoring histograms.

  The overal result as well as the result of the subpackages are stored
  in "status" histograms (see below).

  In addition there are detailed subpackage histograms in order to 
  identify the problem cause. Depending on the "debug level" the number
  of such histograms can be increased to allow detailed study. 
    

  The nomenclature for the histogram IDs is coded in the inline 
  function 

            OtrMonHistId(int level, int algo, int slid, int hid)
  
  where level= LVL_RAW, LVL_HIT, LVL_REC as defined 
                                         by C.Padilla & A.Gellrich
        algo = OtrMon algorithm ID (e.g. = OTRMONAlgoFed)  
               the case algo=0 is reserved for the summary of all
               algorithms
        slid = is the superlayer id, 
               the case slid=0 is reserved for all summary 
               histograms
        hid  = histogram id used by the subpackage.


  Reserved Histogram IDs for "Status" histograms:

         OtrMonHistId(level, algo, OTRMONAllLayer, OTRMONHistSum)
                      (with OTRMONAllLayer=0 and OTRMONHistSum=0)  
         is reserved for the "status" summary of the algorithm algo.


         OtrMonHistId(level, OTRMONHistSum, OTRMONAllLayer, OTRMONHistSum)
                      (with OTRMONAllLayer=0 and OTRMONHistSum=0)  
         is reserved for the "status" summary of all algorithms.
 
  Status Histograms:
    
            32 Bins: 0. -- 32.          
            Every Bin corresponds to a chamber (Superlayer ID)
 
 */
#include "component.h"
#include "grr.h"
#include "etb.h"
#include "dmon.h"
#include "par.h"
#include "smc.h"
#include "rhp.h"
#include "Otr/OtrI.h"
#include "Otr/OtrMon.h"
#include "Otr/OtrDq.h"
#include "Otr/OtrEvent.h"

/*------------------------------------------------------------------------
  For debugging puposes it might be convinient to copy RHP histos to 
  HBOOK histos and storing them to file
  ------------------------------------------------------------------------*/

/* #define USEHBOOK */
#ifdef USEHBOOK
#include "faHbook.h"
#include "rhp2hbook.h"
#define PAWC_SIZE 200000
#define LUNHBK 1
#ifdef extname
float pawc_[PAWC_SIZE];
#else
float pawc_[PAWC_SIZE];
#endif
#define CHDIR "OUTPUT"
#define CHNAME "otrmon.hbook"

#endif 
/*-------------------------------------------------------------------------*/


/* Fortran Interface */
#ifdef __cplusplus
extern "C" {
#endif
  void  OtrMonSwitch_(int* i)        { OtrMonSwitch(*i); }
  int   OtrMonInitialize_()          { return OtrMonInitialize(); }
  int   OtrMonReinitialize_()        { return OtrMonUpdate(); }
  int   OtrMonWriteToDb_()           { return OtrMonWriteToDb(); }
  int   OtrMonClose_()               { return OtrMonClose(); }
  int   OtrMonScan_()                { return OtrMonScan(0); }
  int   OtrMonPrint_()               { return OtrMonPrint(0); }
  int   OtrMonTestMe_()              { return OtrMonTestMe(); }

#ifdef __cplusplus
  }
#endif


#define GOOD 1
#define DEAD 2
#define NOISY 3
#define SL_All 1
#define SL_5mm 2
#define SL_10mm 3

#define FedHistId(A, B)     OtrMonHistId(LVL_RAW,OTRMONAlgoFed, (A), (B))
#define StatusHistId(A, B)  OtrMonHistId(LVL_RAW,OTRMONAlgoStatus, (A), (B))
#define OccupHistId(A, B)   OtrMonHistId(LVL_RAW,OTRMONAlgoOccup, (A), (B))
#define SynchrHistId(A, B)  OtrMonHistId(LVL_RAW,OTRMONAlgoSynchr, (A), (B))
  
static int OtrMonAlgorithm=OTRIInvalid;
static int OtrMonInitialization=0;
static int OtrMonDebugLevel = OTRDQNodebug;
static int OtrMonEvents=0;
static float OtrMonTargetRate=0.;
static int OtrMonOccupEventCounter=0;

static int OtrMonFedLevel      = 1;
static int OtrMonStatusLevel   = 1;
static int OtrMonOccupLevel    = 1;
static int OtrMonSynchrLevel   = 1;
//static int OtrMonTimeLevel     = 1;
static int OtrMonFedCycle      = 100;
static int OtrMonStatusCycle   = 100;    
static int OtrMonOccupCycle    = 100;
static int OtrMonSynchrCycle   = 100;
static int OtrMonTimeCycle     = 100;


/* Some internal functions... */
int   OtrMonErrMsg(char *,int,int);
int   OtrMonFedInit();
int   OtrMonOccupInit();
int   OtrMonStatusInit();
int   OtrMonSynchrInit();
int   OtrMonFedUpdate();
int   OtrMonOccupUpdate();
int   OtrMonStatusUpdate();
int   OtrMonSynchrUpdate();
int   OtrMonTimeUpdate();  
int   OtrMonFedEvent();
int   OtrMonOccupEvent();
int   OtrMonStatusEvent();
int   OtrMonSynchrEvent();
int   OtrMonTimeEvent();
int   OtrMonFedBegin();
int   OtrMonOccupBegin();
int   OtrMonStatusBegin();
int   OtrMonSynchrBegin();
int   OtrMonTimeBegin();
int   OtrMonFedEnd();
int   OtrMonOccupEnd();
int   OtrMonStatusEnd(); 
int   OtrMonSynchrEnd();
int   OtrMonTimeEnd();
int   OtrMonFedHalt();
int   OtrMonOccupHalt();
int   OtrMonStatusHalt();
int   OtrMonSynchrHalt();
int   OtrMonTimeHalt();
int  OtrMonHbookCopyandStore();
void OtrMonSlayerInfoClear();
void OtrMonSlayerInfoEvent();
int OtrISlFromFedId(int fedid);
int OtrISlFromAsdIndex(int asd);
void OtrMonAddAndResetHist(int from, int to);
void OtrMonCopyAndResetHist(int from, int to);

void DivideHistByHist(int hist_to, int hist_nom, int hist_denom, 
		      float scale, int nbins);
void DivideHistByVect(int hist_to, int hist_nom, float *vect, 
		      float scale, int nbins);
void DivideVectByVect(int hist_to, float *vect1, float *vect2, 
		      float scale, int nbins);
void OtrMonSlayerInfoEvent();
int OtrMonSlayerInfoHits(int flag, int sl);
int OtrMonSlayerInfoTsum(int flag, int sl);
int OtrMonSlayerInfoT2sum(int flag, int sl);
int OtrMonStatusEval(float occup);
int OtrMonFedEval(float errors);
int OtrMonStatusEval(float dead, float noise, int sl);
void OtrMonFillSummary(int hid);

/*-----------------------------------------------------------------------*/

int OtrISlFromFedId(int fedid) {
  int i=0;
  int ci=OtrIFromFed(fedid&0xffff,0);
  while (ci<=0 && i<128) {
    ci=OtrIFromFed(fedid&0xffff,i++);
  }
  if(ci>0) {
    return OtrISuperLayer(ci);
  } 
  return -1;
}
int OtrISlFromAsdIndex(int asd) {
  int i=0;
  int ci=OtrIFromAsd(asd,0);
  while (ci<=0 && i<128) {
    ci=OtrIFromAsd(asd,i++);
  }
  if(ci>0) {
    return OtrISuperLayer(ci);
  } 
  return -1;
}

void OtrMonAddAndResetHist(int from, int to) {
  float tmp[2000];
  rhp_hunpak(from, tmp);
  rhp_reset(from);
  rhp_hpak(to, tmp);
}
void OtrMonCopyAndResetHist(int from, int to) {
  rhp_reset(to);
  OtrMonAddAndResetHist(from, to);
}


void DivideHistByHist(int hist_to, int hist_nom, int hist_denom, 
		      float scale, int nbins) {
  float result[32];
  float nominator[32];
  float denominator[32];
  int i;
  rhp_reset(hist_to);
  if (nbins<=32) {
    rhp_hunpak(hist_nom, nominator);
    rhp_hunpak(hist_denom, denominator);
    for(i=0; i<nbins; i++) {
      if(denominator[i]>0) result[i]=scale*nominator[i]/denominator[i];
      else result[i]=-1.;
    }
    rhp_hpak(hist_to, result);
  } else { 
    printf("HistbyHist> Too many bins\n");
  }
}
void DivideHistByVect(int hist_to, int hist_nom, float *vect, 
		      float scale, int nbins) {
  float result[32];
  float nominator[32];
  int i;
  rhp_reset(hist_to);
  if (nbins<=32) {
    rhp_hunpak(hist_nom, nominator);
    for(i=0; i<nbins; i++) {
      if(vect[i]>0) result[i]=scale*nominator[i]/vect[i];
      else result[i]=-1.;
    }
    rhp_hpak(hist_to, result);
  } else { 
    printf("HistbyVect> Too many bins\n");
  }
}
void DivideVectByVect(int hist_to, float *vect1, float *vect2, 
		      float scale, int nbins) {
  float result[32];
  int i;
  rhp_reset(hist_to);
  if (nbins<=32) {
    for(i=0; i<nbins; i++) {
      if(vect2[i]>0) result[i]=scale*vect1[i]/vect2[i];
      else result[i]=-1.;
    }
    rhp_hpak(hist_to, result);
  } else { 
    printf("VectbyVect> Too many bins\n");
  }
}


/**
  Tasks:
  - switch between different solutions of the task of the package
  - also switch between the monitor levels
  */
void  OtrMonSwitch(int i) {
  OtrMonAlgorithm=i;
}


/**
  Tasks:
    - Memory allocation 
    - prepare for reinitialization
    - after this call reinitalization
    
  All function of the package (beside OtrMonSetAlgorithm) have to check, 
  if the initialization was done properly.
  */
  
int OtrMonInitialize(){
  char str[100];

  sprintf(str, "OtrMonInitialize> Start Initilization");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_0); 

  if (OtrMonAlgorithm==OTRIInvalid) OtrMonAlgorithm=0;
  if (OtrMonAlgorithm==0) {
      sprintf(str, "OtrMonInitialize> Nothing selected->No Intit done\n");
      OtrMonErrMsg(str, OTRMONMsgWarn, OTRMONDebug_0); 
      OtrMonInitialization=0;
      return OTRIValid;
  }

  par_book("Events", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonEvents);

  if (OtrMonAlgorithm&OTRMONDebug) OtrMonDebugLevel=OTRMONDebug_1;
  par_book("OtrDebug", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonDebugLevel);

  if (OtrMonAlgorithm&OTRMONAlgoFed) {
    if (OTRMONValid==OtrMonFedInit()) {
      OtrMonInitialization = OtrMonInitialization|OTRMONAlgoFed;
      if (OtrMonAlgorithm&OTRMONAlgoFedDbug) OtrMonFedLevel=2;
    } else {
      OtrMonFedLevel=0;
    }
    par_book("FedLevel", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonFedLevel);
    par_book("FedCycle", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonFedCycle);
  }
  if (OtrMonAlgorithm&OTRMONAlgoStatus) {
    if (OTRMONValid==OtrMonStatusInit()) {
      OtrMonInitialization = OtrMonInitialization|OTRMONAlgoStatus;
      if (OtrMonAlgorithm&OTRMONAlgoStatusDbug) OtrMonStatusLevel=2;
    } else {
      OtrMonStatusLevel=0;
    }
    par_book("StatusLevel", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonStatusLevel);
    par_book("StatusCycle", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonStatusCycle);
  }
  if (OtrMonAlgorithm&OTRMONAlgoOccup) {
    if (OTRMONValid==OtrMonOccupInit()) {
      OtrMonInitialization = OtrMonInitialization|OTRMONAlgoOccup;
      if (OtrMonAlgorithm&OTRMONAlgoOccupDbug) OtrMonOccupLevel=2;
    } else {
      OtrMonOccupLevel=0;
    }
    par_book("OccupLevel", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonOccupLevel);
    par_book("OccupCycle", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonOccupCycle);
  }
  if (OtrMonAlgorithm&OTRMONAlgoSynchr) {
    if (OTRMONValid==OtrMonSynchrInit()) {
      OtrMonInitialization = OtrMonInitialization|OTRMONAlgoSynchr;
      if (OtrMonAlgorithm&OTRMONAlgoSynchrDbug) OtrMonSynchrLevel=2;
    } else {
      OtrMonSynchrLevel=0;
    }
    par_book("SynchrLevel", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonSynchrLevel);
    par_book("SynchrCycle", PAR_T_INT|PAR_T_PUBLISH,1,&OtrMonSynchrCycle);
  }

  if ( ((OtrMonAlgorithm&0xff00ff)&OtrMonInitialization)
       == (OtrMonAlgorithm&0xff00ff)) {
    sprintf(str, "OtrMonInitialize> successful initialized.");
    OtrMonErrMsg(str, OTRMONMsgInfo, OTRMONDebug_0); 
    return OTRIValid;
  }
  sprintf(str, "OtrMonInitialize> failed.");
  OtrMonErrMsg(str, OTRMONMsgWarn, OTRMONDebug_0); 
  return OTRIInvalid;
}



/**
  Tasks:
    - update memory structures, arrays ... from database or ascii files 
  */
int OtrMonUpdate(){
  if(OtrMonAlgorithm&OTRMONAlgoFed) OtrMonFedUpdate(); 
  if(OtrMonAlgorithm&OTRMONAlgoOccup) OtrMonOccupUpdate(); 
  if(OtrMonAlgorithm&OTRMONAlgoStatus) OtrMonStatusUpdate(); 
  if(OtrMonAlgorithm&OTRMONAlgoSynchr) OtrMonSynchrUpdate(); 
  if(OtrMonAlgorithm&OTRMONAlgoTime) OtrMonTimeUpdate(); 
  return 1;
}
  
/**
  Tasks:
    - provide output memory structures, arrays ... to database or ascii files 
  */
int OtrMonWriteToDb(){ 
  if(OtrMonAlgorithm==OTRIInvalid) return OTRIInvalid;
  return OTRIInvalid;
}

/**
  Tasks:
    - free allocate memory
    - clean up to be in a state that initialization is necessary 
  */
  
int   OtrMonClose()     { 
  if (OtrMonAlgorithm==OTRIInvalid) return OTRIInvalid;
  return OTRIInvalid;
  }

/**
  Tasks:
    - scan ascii files into memory structures.
    - if fp==0 assume (Fortran case), try to take filename from environment 
  */
int   OtrMonScan(FILE *fp) { 
  if (OtrMonAlgorithm==OTRIInvalid) return OTRIInvalid;
  if (fp) fprintf(fp, "OtrMonScan> Nothing defined\n");
  return OTRIInvalid;
  }

/**
  Tasks:
    - print memory structure to ascii files
    - if fp==0 assume (Fortran case), try to take filename from environment 
  */
  
int OtrMonPrint(FILE *fp){ 
  if (OtrMonAlgorithm==OTRIInvalid) return OTRIInvalid;
  if (fp) fprintf(fp, "OtrMonPrint> Nothing defined\n");
  return OTRIInvalid;
}

/**
  Tasks:
    - test bench to be sucessfully executed before any release of a new version
    - it is requested, that the author carefully thinks about testing
    - in addition it should be a collection how to test bugs, which happened
      at least once 
  */

int OtrMonTestMe(){
  if (OtrMonAlgorithm==OTRIInvalid) return OTRIInvalid;
  return OTRIInvalid;
}
  
  
int OtrMonErrMsg(char *str, int msg_type, int dbg_lvl){ 
  if (msg_type==GRR_SINFO) {
    if (dbg_lvl<=OtrMonDebugLevel) {
      if (OtrMonAlgorithm&OTRMONOnline) 
	grr_error(GRR_SDEBUG,str,msg_type);
      else printf("%s\n", str);}}
  else { 
    if (OtrMonAlgorithm&OTRMONOnline) 
      grr_error(GRR_SDEBUG,str,msg_type);
    else printf("%s\n", str);
  }
  return 1;
}


int OtrMonEvent(){
  OtrMonEvents++;
  OtrMonSlayerInfoClear();
  OtrMonSlayerInfoEvent();

  if (OtrMonAlgorithm&OTRMONAlgoFed) {
    OtrMonFedEvent();
    if((OtrMonEvents%OtrMonFedCycle)==0 && (OtrMonAlgorithm&OTRMONOnline)) { 
      OtrMonFedUpdate();}
  } 
  if (OtrMonAlgorithm&OTRMONAlgoOccup) {
    OtrMonOccupEvent(); 
    if((OtrMonEvents%OtrMonOccupCycle)==0 && (OtrMonAlgorithm&OTRMONOnline)) {
      OtrMonOccupUpdate();}
  }
  if (OtrMonAlgorithm&OTRMONAlgoStatus) {
    OtrMonStatusEvent(); 
    if((OtrMonEvents%OtrMonStatusCycle)==0 && (OtrMonAlgorithm&OTRMONOnline)) {
      OtrMonStatusUpdate();}
  }
  if (OtrMonAlgorithm&OTRMONAlgoSynchr) {
    OtrMonSynchrEvent();
    if((OtrMonEvents%OtrMonSynchrCycle)==0 && (OtrMonAlgorithm&OTRMONOnline))
      OtrMonSynchrUpdate();
  }
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoTime) {
    OtrMonTimeEvent(); 
    if((OtrMonEvents%OtrMonTimeCycle)==0 && OtrMonAlgorithm&OTRMONOnline) 
      OtrMonTimeUpdate();
  }
  return 1;
}

int OtrMonBegin(){
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoFed)    OtrMonFedBegin(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoOccup)  OtrMonOccupBegin(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoStatus) OtrMonStatusBegin(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoSynchr) OtrMonSynchrBegin(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoTime)   OtrMonTimeBegin(); 
  return 1;
}

int OtrMonEnd(){
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoFed)    OtrMonFedEnd(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoOccup)  OtrMonOccupEnd(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoStatus) OtrMonStatusEnd(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoSynchr) OtrMonSynchrEnd(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoTime)   OtrMonTimeEnd(); 
  return 1;
}

int OtrMonHalt(){
  /* if (!(OtrMonAlgorithm&OTRMONOnline)) OtrMonUpdate(); */
  OtrMonUpdate(); 

  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoFed)    OtrMonFedHalt(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoOccup)  OtrMonOccupHalt(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoStatus) OtrMonStatusHalt(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoSynchr) OtrMonSynchrHalt(); 
  if (OtrMonAlgorithm&OtrMonInitialization&OTRMONAlgoTime)   OtrMonTimeHalt(); 

  /*  OtrMonHbookCopyandStore();*/
#ifdef USEHBOOK
  OtrMonHbookCopyandStore();
#endif
  return 1;
}

int OtrMonHelp(int argc, const char **argv){
  int i;
  for(i=1;i<argc;i++){
    if(      strcmp(argv[i],"-debug") == 0) printf("No help\n");
    else if (strcmp(argv[i],"-junk") == 0)  printf("No help\n"); 
  }
  return OTRIValid;
}

/*--------------------------------------------------------------------

              S U B - P A C K A G E    P A R T 

  --------------------------------------------------------------------*/


/* -----------------  FED  Package  ----------------------------------
   -------------------------------------------------------------------*/
#define ALL 0
#define FEDSummaryId FedHistId(ALL, 0)
#define FEDBadId     FedHistId(ALL, 1) 
#define FEDFedSlId     FedHistId(ALL, 2) 
#define FEDBadDetailsId FedHistId(ALL, 11)  
#define FEDBadDetailsAllId FedHistId(ALL, 12)  

/*  Histograms: 
    FEDSummaryId: Summary, 1 Bin for every chamber (Good/Usable/Bad)
    FEDBadId    : Histogram with the total number of bad FEDs/SL in all events
    FEDFedSlId    : Histogram telling, how many FEDs there are per SL
    FEDBadFedsId: Histogram with the number of bad FEDs/SL per event
    FEDBadDetailsId: Histogram (1 channel/FED) with numb of problems/FED
    FEDBadDetailsAllId: Histogram (1 channel/FED) with numb of problems/FED
                        of all events 
    */

int  OtrMonFedInit(){

  char str[80];
  sprintf(str, "OtrMonFedInit> Start Initilization");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_0); 

  rhp_hbook1(FEDSummaryId,"FED Summary",
	     OTRMONHistBins,0.,(float)OTRMONHistBins);
  rhp_hbook1(FEDBadId,"FED Bad",
	     OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(FEDFedSlId,"FED Fed/Sl",
	     OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(FEDBadDetailsId,"FED Bad Details",
	     OtrIFedNumberOf(),0.,(float)OtrIFedNumberOf());
  rhp_hbook1(FEDBadDetailsAllId,"FED Bad Details All", 
	     OtrIFedNumberOf(),0.,(float)OtrIFedNumberOf());
  return OTRIValid;
}

int OtrMonFedBegin(){
  int i, sl, fedid, mcn, mcp;
  rhp_reset(FEDSummaryId);
  rhp_reset(FEDBadId);
  rhp_reset(FEDFedSlId);
  rhp_reset(FEDBadDetailsId);
  rhp_reset(FEDBadDetailsAllId);

  for (i=0; i<OtrIFedNumberOf(); i++) {
    fedid=OtrIFedIdFromIndex(i);
    sl = OtrISlFromFedId(fedid);
    if (sl<=8 && sl>0) mcn=1;
    else mcn=0;
    if (sl<=24 && sl>16) mcp=1;
    else mcp=0;
    rhp_hfill(FEDFedSlId, (float) sl, 0., 1.);
    if (mcn) rhp_hfill(FEDFedSlId, 7., 0., 1.);
    if (mcp) rhp_hfill(FEDFedSlId, 23., 0., 1.);
  }
  return 1;
}


int OtrMonFedEvent(){
  int i,sl,fedid, mcp, mcn;
  for (i=0; i<OtrIFedNumberOf(); i++) {
    fedid=OtrIFedIdFromIndex(i);
    sl = OtrISlFromFedId(fedid);
    if (sl<=8 && sl>0) mcn=1;
    else mcn=0;
    if (sl<=24 && sl>16) mcp=1;
    else mcp=0;
    if (!OtrEventFedStatus(i))  {
      rhp_hfill(FEDBadId, (float) sl, 0., 1.);
      if (mcn) rhp_hfill(FEDBadId, 7., 0., 1.);
      if (mcp) rhp_hfill(FEDBadId, 23., 0., 1.);
      rhp_hfill(FEDBadDetailsAllId, (float) i, 0., 1.);
      rhp_hfill(FEDBadDetailsId, (float) i, 0., 1.);
    }
  }
  return 1;
}

/* Cristobals Histograms should not be pubished - he will get them */
int OtrMonFedUpdate(){
  float bad[OTRMONHistBins];
  float all[OTRMONHistBins];
  float fed[OTRMONHistBins];
  int i;
  char str[80];

  sprintf(str, "OtrMonFedUpdate> Updating");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_1); 

  rhp_reset(FEDSummaryId);

  rhp_hunpak(FEDBadId, bad);
  rhp_hunpak(FEDFedSlId, all);
  rhp_reset(FEDBadId);

  for (i=1; i<OTRMONHistBins; i++) {
    if(all[i]>0) {
      fed[i]=bad[i]/all[i]; 
      rhp_hfill(FEDSummaryId, (float) i, 0., (float) OtrMonFedEval(fed[i]));
    } else {
      fed[i]=-1.;
      rhp_hfill(FEDSummaryId, (float) i, 0., (float) OTRMONUnknown);
    }
  }
  fed[0]=-1.;
  OtrMonFillSummary(FEDSummaryId);
  rhp_hpak(FEDBadId, fed);

  if(OtrMonFedLevel==2) rhp_publish(FEDBadDetailsId);

  /* reset if necessary */
  rhp_reset(FEDBadDetailsId);
  return 1;
}

int OtrMonFedEnd(){
  return 1;
}

int OtrMonFedHalt() {
  return 1;
}

int OtrMonFedEval(float errors){
  if (errors<-0.5) return OTRMONUnknown;
  if (errors<0.05) return OTRMONGood;
  if (errors<0.10) return OTRMONUsable;
  else return OTRMONBad;
  return OTRMONUnknown;
} 

void OtrMonFedMyHistIds(int *nb, int *ids){
  *nb=5;
  ids[0]=FEDSummaryId;
  ids[1]=FEDBadId;
  ids[2]=FEDFedSlId;
  ids[3]=FEDBadDetailsId;
  ids[4]=FEDBadDetailsAllId;
}

/* ----------------- Status Evaluation Package -----------------------
   ------------------------------------------------------------------*/

/* purpose:
   -  monitor the number of dead/noisy channels for different SL
      The way it is programmed assumes that OtrDq regularly updates 
      the status word
   */

#define STATUSSummaryId StatusHistId(ALL, 0)
#define STATUSDeadId    StatusHistId(ALL, 1) 
#define STATUSNoisyId   StatusHistId(ALL, 2) 
#define STATUSDeadHitsId    StatusHistId(ALL, 3) 
#define STATUSNoisyHitsId   StatusHistId(ALL, 4) 
#define STATUSDeadOccupId    StatusHistId(ALL, 7) 
#define STATUSNoisyOccupId   StatusHistId(ALL, 8) 
#define STATUSNoisy5Id    StatusHistId(ALL, 9) 
#define STATUSNoisy10Id   StatusHistId(ALL, 10) 

int OtrMonStatusInit(){
  char str[80];
  sprintf(str, "OtrMonStatusInit> Start Initilization");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_0); 

  rhp_hbook1(STATUSSummaryId,"STATUS Summary",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(STATUSDeadId,"STATUS Dead channels",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(STATUSDeadHitsId,"STATUS Dead channels hit",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(STATUSDeadOccupId,"STATUS Dead channels occup",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(STATUSNoisyId,"STATUS Noisy channel",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(STATUSNoisyHitsId,"STATUS Noisy channels hit",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(STATUSNoisyOccupId,"STATUS Noisy channels occup",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(STATUSNoisy5Id,"STATUS Noisy channel",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(STATUSNoisy10Id,"STATUS Noisy channel",
	   OTRMONHistBins,0.,(float) OTRMONHistBins);
  return 0;
}

int OtrMonStatusBegin(){
  rhp_reset(STATUSSummaryId);
  rhp_reset(STATUSDeadId);
  rhp_reset(STATUSNoisyId);
  rhp_reset(STATUSDeadHitsId);
  rhp_reset(STATUSNoisyHitsId);
  rhp_reset(STATUSDeadOccupId);
  rhp_reset(STATUSNoisyOccupId);
  rhp_reset(STATUSNoisy5Id);
  rhp_reset(STATUSNoisy10Id);

  return 1;
}
int OtrMonStatusEvent(){
  int i;
  /* ------- loop over all superlayers ---------- */
  for(i=1;i<OTRMONHistBins;i++){ 
    rhp_hfill(STATUSDeadHitsId, ((float)i)+0.5, 0., 
	      (float) OtrMonSlayerInfoHits(DEAD, i));
    rhp_hfill(STATUSNoisyHitsId, ((float)i)+0.5, 0., 
	      (float) OtrMonSlayerInfoHits(NOISY, i));
  }

  rhp_hfill(STATUSDeadHitsId, 0.5, 0., 
	    (float) OtrMonSlayerInfoHits(DEAD, 0));
  rhp_hfill(STATUSNoisyHitsId, 0.5, 0., 
	    (float) OtrMonSlayerInfoHits(NOISY, 0));
  
  rhp_hfill(STATUSDeadHitsId, ((float)OTRMONHistBins)-0.5, 
	    0., 1.);
  rhp_hfill(STATUSNoisyHitsId, ((float)OTRMONHistBins)-0.5, 
	    0., 1.);
  return 1;
}

/*---------------------------------------------------------------*/
int OtrMonStatusUpdate(){
  float ratio_dead, ratio_noisy;
  float dead[OTRMONHistBins];
  float noisy[OTRMONHistBins];
  float all[OTRMONHistBins];
  float dead5[OTRMONHistBins];
  float noisy5[OTRMONHistBins];
  float all5[OTRMONHistBins];
  float dead10[OTRMONHistBins];
  float noisy10[OTRMONHistBins];
  float all10[OTRMONHistBins];

  int i,sl,ci, sec;
  int asd,cha;
  int mcp, mcn;
  float events;
  char str[80];

  sprintf(str, "OtrMonStatusUpdate> Updating");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_1); 

  for(i=0;i<OTRMONHistBins;i++){
    dead[i]=0.; dead5[i]=0.; dead10[i]=0.;
    noisy[i]=0.; noisy5[i]=0.; noisy10[i]=0.;
    all[i]=0.; all5[i]=0.; all10[i]=0.;
  }

  /* loop over all channels and count the bad ones */
  for(asd=0;asd<OtrIAsdNumberOf();asd++){
    for(cha=0;cha<OtrIAsdNumberOfChannel();cha++){
      ci=OtrIFromAsd(asd, cha);
      if (ci>=0) {
	sl=OtrISuperLayer(ci);
	if (sl<=8 && sl>0) mcn=1;
	else mcn=0;
	if (sl<=24 && sl>16) mcp=1;
	else mcp=0;
        sec=OtrISector(ci);
	if(OtrDqDead(ci)) { 
	  dead[0]+=1.; dead[sl]+=1.; 
	  if(mcn)dead[7]+=1.; if(mcp)dead[23]+=1.;}
	else if(OtrDqBad(ci)) { 
	  noisy[0]+=1.; noisy[sl]+=1.;
	  if(mcn)noisy[7]+=1.; if(mcp)noisy[23]+=1.;}
	all[sl]+=1.; all[0]+=1.; if(mcn)all[7]+=1.; if(mcp)all[23]+=1.; 
	if (sec<7) {
	  if(OtrDqDead(ci)) { dead5[0]+=1.; dead5[sl]+=1.; }
	  else if(OtrDqBad(ci)) { noisy5[0]+=1.; noisy5[sl]+=1.;}
	  all5[sl]+=1.; all5[0]+=1.;
	} else {
	  if(OtrDqDead(ci)) { dead10[0]+=1.; dead10[sl]+=1.; }
	  else if(OtrDqBad(ci)) { noisy10[0]+=1.; noisy10[sl]+=1.;}
	  all10[sl]+=1.; all10[0]+=1.;
	}
      } 
    }
  } 

  /* ---------- fill status histograms  ------------------ */
  rhp_reset(STATUSSummaryId);
  rhp_reset(STATUSDeadId);
  rhp_reset(STATUSNoisyId);
  rhp_reset(STATUSNoisy5Id);
  rhp_reset(STATUSNoisy10Id);

  DivideVectByVect(STATUSDeadId,   dead, all, 1., OTRMONHistBins);
  DivideVectByVect(STATUSNoisyId, noisy, all, 1., OTRMONHistBins);
  DivideVectByVect(STATUSNoisy5Id, noisy5, all5, 1., OTRMONHistBins);
  DivideVectByVect(STATUSNoisy10Id, noisy10, all10, 1., OTRMONHistBins);
  
  for(i=1;i<OTRMONHistBins;i++){ 
    if(dead[i]>=0.0 && noisy[i]>=0.0) {
      ratio_dead  =  dead[i]/all[i];
      ratio_noisy = noisy[i]/all[i];
      rhp_hfill(STATUSSummaryId,(float)i, 0.,
		(float) OtrMonStatusEval(ratio_dead, ratio_noisy, i));
    } else {
      rhp_hfill(STATUSSummaryId,(float)i,0.,(float)OTRMONUnknown);
    }
  }
  OtrMonFillSummary(STATUSSummaryId);


  /* --------- fill occup of bad channels -------------- */
  events=rhp_hi(STATUSDeadHitsId, OTRMONHistBins);
  if(events>0) events=1./events;
  else events=-1.;
  DivideHistByVect(STATUSDeadOccupId, STATUSDeadHitsId, dead, 
		   events,OTRMONHistBins);
  DivideHistByVect(STATUSNoisyOccupId, STATUSNoisyHitsId, noisy, 
		   events, OTRMONHistBins);


  /* for (i=0; i<OTRMONHistBins;i++){
     printf("Sl %d -- Noisy Hits: %f Noisy Channels %f \n", 
     i, rhp_hi(STATUSNoisyHitsId, i+1), noisy[i]);
     printf("SL %d -- Dead Hits: %f Dead Channels %f \n", 
     i, rhp_hi(STATUSDeadHitsId, i+1), dead[i]);
     }
     */
  rhp_reset(STATUSDeadHitsId);
  rhp_reset(STATUSNoisyHitsId);


  return 1;
}

int OtrMonStatusEnd(){
  return 1;
}
int OtrMonStatusHalt(){
  return 1;
}

int OtrMonStatusEval(float dead, float noise, int sl) {
  static float dead_ref[32] = {
    0.105001, 0.0480769, 0.0877976, 0.0295139, 0.0910088,
    0.245739, 0.0682292, 0.077,       0.0823413, 0.0761876, 
    0.125904, 0.100151, 0.0787,    0.125535,  0.0648721, 
    -1.,      -1.,      0.108333,  0.111742,  0.247222,  
    0.14388,  0.254573, 0.0349265, 0.13,       0.100643,  
    0.136089, 0.109421, 0.113872,  0.138625,  0.117093,  
    0.101923, -1};
  static float noise_ref[32] = {
    0.029315, 0.00641026, 0.0208333, 0.0121528, 0.00932018,
    0.0142045, 0.0479167, 0.022, 0.0352183, 0.0363325,
    0.0233434, 0.0332831, 0.031, 0.0265411, 0.0334158,
    -1., -1., 0.0177083, 0.00757576, 0.0284722,
    0.00911458, 0.0716463, 0.0122549, 0.019, 0.0110294,
    0.0336442, 0.0313427, 0.0281899, 0.0395, 0.0279869,
    0.0266827, -1};
  
  if (noise<0.001 || dead<0.001) return OTRMONUnknown;
  if ((noise-noise_ref[sl])+(dead-dead_ref[sl])<0.05) return OTRMONGood;
  if ((noise-noise_ref[sl])+(dead-dead_ref[sl])<0.10) return OTRMONUsable;
  else  return OTRMONBad;
  return OTRMONUnknown;
}

void OtrMonStatusMyHistIds(int *nb, int *ids){
  *nb=7;
  ids[0]= STATUSSummaryId;
  ids[1]= STATUSDeadId; 
  ids[2]= STATUSNoisyId;
  ids[3]= STATUSDeadOccupId;
  ids[4]= STATUSNoisyOccupId;
  ids[5]= STATUSNoisy5Id;
  ids[6]= STATUSNoisy10Id;
}

/* -----------------  Occupancy  Package  -----------------------------
   -------------------------------------------------------------------*/

/* purpose:
   - get occupancy (dead/noisy channels suppressed) for every SL
   - calculate a "rate equivalent"
   - trace the total number of hits (no noise suppression!) over time
   */
#define ALL 0
#define OCCUPSummaryId OccupHistId(ALL, 0)
#define OCCUPHitsId   OccupHistId(ALL, 1) 
#define OCCUPOccupId   OccupHistId(ALL, 2) 
#define OCCUPAllChanId   OccupHistId(ALL, 3) 
#define OCCUPEventOccup   OccupHistId(ALL, 4) 
#define OCCUPEventOccupTmp OCCUPEventOccup+1
#define OCCUPEventGoodOccup    OccupHistId(ALL, 12) 
#define OCCUPEventGoodOccupTmp OccupHistId(ALL, 13) 
#define OCCUPGoodHitsId   OccupHistId(ALL, 10) 
#define OCCUPGoodOccupId   OccupHistId(ALL, 11) 


int OtrMonOccupInit() {
  char str[80];
  sprintf(str, "OtrMonOccupInit> Start Initilization");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_0); 

  rhp_hbook1(OCCUPSummaryId,"OCCUP summary",
	     OTRMONHistBins,0.,(float)OTRMONHistBins);
  rhp_hbook1(OCCUPOccupId,"OCCUP occup per SL",
	     OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(OCCUPHitsId,"OCCUP hits per SL",
	     OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(OCCUPAllChanId,"OCCUP all channels in SL",
	     OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(OCCUPEventOccupTmp,"OCCUP Hits in Event",
	     30, 0., 30000.);
  rhp_hbook1(OCCUPEventOccup,"OCCUP Hits in Event",
	     30, 0., 30000.);
  rhp_hbook1(OCCUPEventGoodOccupTmp,"OCCUP Hits in Event",
	     30, 0., 30000.);
  rhp_hbook1(OCCUPEventGoodOccup,"OCCUP Hits in Event",
	     30, 0., 30000.);

  rhp_hbook1(OCCUPGoodOccupId,"OCCUP good chan occup per SL",
	     OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(OCCUPGoodHitsId,"OCCUP good chan hits per SL",
	     OTRMONHistBins,0.,(float) OTRMONHistBins);

  return OTRIValid;
}
int OtrMonOccupBegin() {
  int i, sl, mcn, mcp;
  rhp_reset(OCCUPSummaryId);
  rhp_reset(OCCUPOccupId);
  rhp_reset(OCCUPHitsId);
  rhp_reset(OCCUPAllChanId);
  rhp_reset(OCCUPEventOccup);
  rhp_reset(OCCUPEventGoodOccup);
  rhp_reset(OCCUPGoodOccupId);
  rhp_reset(OCCUPGoodHitsId);

  for (i=0; i<OtrIAsdNumberOf(); i++) {
    sl=OtrISlFromAsdIndex(i);
    if (sl<=8 && sl>0) mcn=1;
    else mcn=0;
    if (sl<=24 && sl>16) mcp=1;
    else mcp=0;
    rhp_hfill(OCCUPAllChanId, (float) sl, 0., 16.);
    if (mcn) rhp_hfill(OCCUPAllChanId, 7., 0., 16.);
    if (mcp) rhp_hfill(OCCUPAllChanId, 23., 0., 16.);
  }
  return 1;
}
int OtrMonOccupEvent() {
  int i;

  OtrMonTargetRate+=OtrEventInfoIaRate();
  OtrMonOccupEventCounter++;

  if(OtrEventStatisticsHits(0)<30000.) 
    rhp_hfill(OCCUPEventOccupTmp, (float)OtrEventStatisticsHits(0), 0., 1.);
  else 
     rhp_hfill(OCCUPEventOccupTmp, 29999. , 0., 1.);   

  if(OtrMonSlayerInfoHits(GOOD, 0)<30000.) 
    rhp_hfill(OCCUPEventGoodOccupTmp, (float)OtrMonSlayerInfoHits(GOOD, 0), 
	      0., 1.);
  else 
     rhp_hfill(OCCUPEventGoodOccupTmp, 29999. , 0., 1.);   

  rhp_hfill(OCCUPHitsId, 0., 0., 1.);
  for (i=1; i<OTRMONHistBins; i++) {
    rhp_hfill(OCCUPHitsId, (float) i, 0., (float) OtrEventStatisticsHits(i));
  }
  rhp_hfill(OCCUPHitsId, (float) OTRMONHistBins, 0., 1.);

  for (i=0; i<OTRMONHistBins; i++) {
    rhp_hfill(OCCUPGoodHitsId, ((float) i)+0.5, 0., 
	     (float)OtrMonSlayerInfoHits(GOOD, i));
  }
  return 1;
}
int OtrMonOccupUpdate() {
  float nhi[OTRMONHistBins];
  float hit[OTRMONHistBins];
  float all[OTRMONHistBins];
  float occ[OTRMONHistBins];
  float noc[OTRMONHistBins];
  int i;
  char str[80];

  sprintf(str, "OtrMonOccupUpdate> Updating");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_1); 

  /* calculate the mean target rate w/r to the last update cycle */
  if(OtrMonOccupEventCounter>0) OtrMonTargetRate=
				  OtrMonTargetRate/((float) OtrMonOccupEventCounter);
  else OtrMonTargetRate=-1.;
  if (OtrMonTargetRate<=0.001) OtrMonTargetRate=-1.;
  /* printf("OtrMonOccupUpdate> Target Rate %f \n", OtrMonTargetRate); */

  rhp_reset(OCCUPSummaryId);
  rhp_reset(OCCUPOccupId);
  rhp_reset(OCCUPGoodOccupId);
  rhp_hunpak(OCCUPHitsId, hit);
  rhp_hunpak(OCCUPGoodHitsId, nhi);
  rhp_hunpak(OCCUPAllChanId, all);
  rhp_reset(OCCUPHitsId);
  rhp_reset(OCCUPGoodHitsId);

  for (i=1; i<OTRMONHistBins; i++) {
    if(all[i]>0&&hit[0]>0) {
      occ[i]=hit[i]/all[i]/hit[0]; 
      noc[i]=nhi[i]/all[i]/hit[0]; 
      rhp_hfill(OCCUPSummaryId, (float) i, 0., (float) OtrMonStatusEval(noc[i]));
    } else {
      occ[i]=-1.;
      noc[i]=-1.;
      rhp_hfill(OCCUPSummaryId, (float) i, 0., (float) OTRMONUnknown);
    }
  }
  rhp_hpak(OCCUPOccupId, occ);
  rhp_hpak(OCCUPGoodOccupId, noc);
  OtrMonFillSummary(OCCUPSummaryId);

  OtrMonCopyAndResetHist(OCCUPEventOccupTmp, OCCUPEventOccup);
  OtrMonCopyAndResetHist(OCCUPEventGoodOccupTmp, OCCUPEventGoodOccup);
  OtrMonTargetRate=0.;
  OtrMonOccupEventCounter=0;

  return 1;
}
int OtrMonOccupEnd() {
  return 1;
}
int OtrMonOccupHalt() {
  return 1;
}

int OtrMonStatusEval(float occup) {
  float occ = occup*10./OtrMonTargetRate;
  if (occ<0.0) return OTRMONUnknown;
  if (occ<.04)  return OTRMONGood;
  if (occ<.08)  return OTRMONUsable;
  else return OTRMONBad;
  return OTRMONUnknown;
}

void OtrMonOccupMyHistIds(int *nb, int *ids){
  *nb=7;
  ids[0]=OCCUPSummaryId;
  ids[1]=OCCUPHitsId;
  ids[2]=OCCUPOccupId;
  ids[3]=OCCUPAllChanId;
  ids[4]=OCCUPEventOccup;
  ids[5]=OCCUPGoodOccupId;
  ids[6]=OCCUPEventGoodOccup;
}

/* -----------------  Synchronization Package  -----------------------
   -------------------------------------------------------------------*/

/* purpose:
   - checks the synchronization of all FED crates (for good channels only):
   Mean/Rms of 5mm, Mean/Rms of 10 mm, BX Occupancy:
   */

#define HERABX 220
#define SYNCHRSummaryId SynchrHistId(ALL, 0)
#define SYNCHRBxPlot    SynchrHistId(ALL, 1)
#define SYNCHREventTime SynchrHistId(ALL, 3)
#define SYNCHREventTime2 SynchrHistId(ALL, 5)
#define SYNCHRTime5mm   SynchrHistId(ALL, 7)
#define SYNCHRTime10mm  SynchrHistId(ALL, 9)

#define SYNCHRBxPlotTmp     (SYNCHRBxPlot+1)
#define SYNCHREventTimeTmp  (SYNCHREventTime+1)
#define SYNCHREventTime2Tmp (SYNCHREventTime2+1)
#define SYNCHRTime5mmTmp    (SYNCHRTime5mm+1)
#define SYNCHRTime10mmTmp   (SYNCHRTime10mm+1)


int OtrMonSynchrInit() {
  char str[80];
  sprintf(str, "OtrMonSynchrInit> Start Initilization");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_0); 
  
  rhp_hbook1(SYNCHRSummaryId,"SYNCHR Summary",
	     OTRMONHistBins,0.,(float) OTRMONHistBins);
  rhp_hbook1(SYNCHRBxPlot, "SYNCHR BX Plot", HERABX, 0., (float) HERABX);
  rhp_hbook1(SYNCHREventTime, "SYNCHR Event Time", 128, 0., 256.);
  rhp_hbook1(SYNCHREventTime2, "SYNCHR Event Time many hits", 128, 0., 256.);
  rhp_hbook1(SYNCHRTime5mm, "SYNCHR TDC all 5mm", 128, 0., 256.);
  rhp_hbook1(SYNCHRTime10mm, "SYNCHR TDC all 10mm", 128, 0., 256.);

  rhp_hbook1(SYNCHRBxPlotTmp, "SYNCHR BX Plot", HERABX, 0., (float) HERABX);
  rhp_hbook1(SYNCHREventTimeTmp, "SYNCHR Event Time", 128, 0., 256.);
  rhp_hbook1(SYNCHREventTime2Tmp, "SYNCHR Event Time many hits", 128, 0., 256.);
  rhp_hbook1(SYNCHRTime5mmTmp,"SYNCHR TDC all 5mm",  128, 0., 256.);
  rhp_hbook1(SYNCHRTime10mmTmp,"SYNCHR TDC all 10mm", 128, 0., 256.);

return OTRIValid;
}
int OtrMonSynchrBegin() {
  rhp_reset(SYNCHRSummaryId);
  rhp_reset(SYNCHRBxPlot);
  rhp_reset(SYNCHREventTime);
  rhp_reset(SYNCHREventTime2);
  rhp_reset(SYNCHRTime5mm);
  rhp_reset(SYNCHRTime10mm);
  rhp_reset(SYNCHRBxPlotTmp);
  rhp_reset(SYNCHREventTimeTmp);
  rhp_reset(SYNCHREventTime2Tmp);
  rhp_reset(SYNCHRTime5mmTmp);
  rhp_reset(SYNCHRTime10mmTmp);
  return 1;
}
int OtrMonSynchrEvent() {
  int i;

  rhp_hfill(SYNCHREventTimeTmp, OtrEventStatisticsTmean(0), 0., 1.);
  if (OtrEventStatisticsHits(0)>2000.) {
    rhp_hfill(SYNCHRBxPlotTmp, (float) OtrEventInfoPhysBx(), 
	      0., (float) OtrEventStatisticsHits(0));
    rhp_hfill(SYNCHREventTime2Tmp, OtrEventStatisticsTmean(0), 0., 1.);
  }

  for (i=0; i<OtrEventHitNumberOf(); i++) {
    if(OtrDqGood(OtrEventHitChannelIndex(i))) {
      if (OtrISector(OtrEventHitChannelIndex(i))<7) 
	rhp_hfill(SYNCHRTime10mmTmp, (float) OtrEventHitTdcValue(i), 0., 1.);
      else 
	rhp_hfill(SYNCHRTime5mmTmp, (float) OtrEventHitTdcValue(i), 0., 1.);
    }
  }

  return 1;
}
int OtrMonSynchrUpdate() {
  char str[80];

  sprintf(str, "OtrMonSynchrUpdate> Updating");
  OtrMonErrMsg(str, GRR_SINFO, OTRMONDebug_1); 

  OtrMonCopyAndResetHist(SYNCHRBxPlotTmp, SYNCHRBxPlot);
  OtrMonCopyAndResetHist(SYNCHREventTimeTmp, SYNCHREventTime);
  OtrMonCopyAndResetHist(SYNCHREventTime2Tmp, SYNCHREventTime2);
  OtrMonCopyAndResetHist(SYNCHRTime10mmTmp, SYNCHRTime10mm);
  OtrMonCopyAndResetHist(SYNCHRTime5mmTmp, SYNCHRTime5mm);
  return 1;
}
int OtrMonSynchrEnd() {
  return 1;
}
int OtrMonSynchrHalt() {
  return 1;
}

void OtrMonSynchrMyHistIds(int *nb, int *ids){
  *nb=6;
  ids[0]=SYNCHRSummaryId;
  ids[1]=SYNCHRBxPlot;
  ids[2]=SYNCHREventTime;
  ids[3]=SYNCHREventTime2;
  ids[4]=SYNCHRTime5mm;
  ids[5]=SYNCHRTime10mm;
}

/* -----------------  Timing  Package -- -----------------------------
   -------------------------------------------------------------------*/

/* purpose:
   - I don't know yet what it should do ???
   */

int OtrMonTimeInit(){
  return 1;
}
int OtrMonTimeBegin(){
  return 1;
}
int OtrMonTimeEvent(){
  return 1;
}
int OtrMonTimeUpdate(){
  return 1;
}
int OtrMonTimeEnd(){
  return 1;
}
int OtrMonTimeHalt(){
  return 1;
}

/*--------------- Make Summary -----------------------------------------*/
void OtrMonFillSummary(int hid){
  int i, nbad=0, nusable=0, nunknown=0, flag=0;
  float tmp[32];
  rhp_hunpak(hid, tmp);
  for (i=7;i<15; i++) {
    if (i!=8) {
      if (((int) tmp[i])==OTRMONUnknown)nunknown++; 
      else if (((int) tmp[i])==OTRMONUsable) nusable++;
      else if (((int) tmp[i])==OTRMONBad) nbad++;
    }
  }
  for (i=23;i<31; i++) {
    if (i!=24) {
      if (((int) tmp[i])==OTRMONUnknown)nunknown++; 
      else if (((int) tmp[i])==OTRMONUsable) nusable++;
      else if (((int) tmp[i])==OTRMONBad) nbad++;
    }
  }
  if      (nunknown>4) flag=OTRMONUnknown;
  else if (nbad>1)     flag=OTRMONBad;
  else if (nusable>2)  flag=OTRMONUsable;
  else                 flag=OTRMONGood;
  rhp_hfill(hid, 0.0, 0., (float) flag);
}

/*--------------- Superlayer Info --------------------------------------
  ---------------------------------------------------------------------*/
static int NumGoodHits[32];
static int Num5mmGoodHits[32];
static int Num10mmGoodHits[32];
static int NumNoisyHits[32];
static int NumDeadHits[32];

static int TsumGoodHits[32];
static int Tsum5mmGoodHits[32];
static int Tsum10mmGoodHits[32];
static int T2sumGoodHits[32];
static int T2sum5mmGoodHits[32];
static int T2sum10mmGoodHits[32];

static float TmeanGoodHits[32];
static float Tmean5mmGoodHits[32];
static float Tmean10mmGoodHits[32];
static float TrmsGoodHits[32];
static float Trms5mmGoodHits[32];
static float Trms10mmGoodHits[32];

 

void OtrMonSlayerInfoClear(){
  memset(NumGoodHits, 0, sizeof(int)*32);  
  memset(Num5mmGoodHits, 0, sizeof(int)*32);  
  memset(Num10mmGoodHits, 0, sizeof(int)*32);  
  memset(NumNoisyHits, 0, sizeof(int)*32);  
  memset(NumDeadHits, 0, sizeof(int)*32);  

  memset(TsumGoodHits, 0, sizeof(int)*32);  
  memset(Tsum5mmGoodHits, 0, sizeof(int)*32);  
  memset(Tsum10mmGoodHits, 0, sizeof(int)*32);  
  memset(T2sumGoodHits, 0, sizeof(int)*32);  
  memset(T2sum5mmGoodHits, 0, sizeof(int)*32);  
  memset(T2sum10mmGoodHits, 0, sizeof(int)*32);  

  memset(TmeanGoodHits, 0, sizeof(float)*32);  
  memset(Tmean5mmGoodHits, 0, sizeof(float)*32);  
  memset(Tmean10mmGoodHits, 0, sizeof(float)*32);  
  memset(TrmsGoodHits, 0, sizeof(float)*32);  
  memset(Trms5mmGoodHits, 0, sizeof(float)*32);  
  memset(Trms10mmGoodHits, 0, sizeof(float)*32);  
}


void OtrMonSlayerInfoEvent(){
  int i, ci, sl, mcp, mcn;
  for (i=0; i<OtrEventHitNumberOf(); i++) {

    ci = OtrEventHitChannelIndex(i);
    sl = OtrISuperLayer(ci);
    if (sl<=8 && sl>0) mcn=1;
    else mcn=0;
    if (sl<=24 && sl>16) mcp=1;
    else mcp=0;

    if(OtrDqGood(ci)) {
      NumGoodHits[0]++;
      NumGoodHits[sl]++;
      if (mcn) NumGoodHits[7]++;
      if (mcp) NumGoodHits[23]++;
      TsumGoodHits[sl]+= OtrEventHitTdcValue(i);
 
      if(OtrISector(ci)<7){
	Num5mmGoodHits[sl]++;
	Tsum5mmGoodHits[sl]+= OtrEventHitTdcValue(i);
      } else {
	Num10mmGoodHits[sl]++;
	Tsum10mmGoodHits[sl]+= OtrEventHitTdcValue(i);}
    } else if (OtrDqDead(ci)) {
      NumDeadHits[0]++;
      NumDeadHits[sl]++;
      if (mcn) NumDeadHits[7]++;
      if (mcp) NumDeadHits[23]++;
    } else {
      NumNoisyHits[0]++;
      NumNoisyHits[sl]++;
      if (mcn) NumNoisyHits[7]++;
      if (mcp) NumNoisyHits[23]++;
    }
  }
}

int OtrMonSlayerInfoHits(int flag, int sl) {
  switch (flag) {
  case (GOOD): return NumGoodHits[sl];
  case (DEAD): return NumDeadHits[sl];
  case (NOISY): return NumNoisyHits[sl];
  }
  return 0;
}
int OtrMonSlayerInfoTsum(int flag, int sl) {
  switch (flag) {
  case (SL_All): return TsumGoodHits[sl];
  case (SL_5mm): return Tsum5mmGoodHits[sl];
  case (SL_10mm): return Tsum10mmGoodHits[sl];
  }
  return 0;
}
int OtrMonSlayerInfoT2sum(int flag, int sl) {
  switch (flag) {
  case (SL_All): return T2sumGoodHits[sl];
  case (SL_5mm): return T2sum5mmGoodHits[sl];
  case (SL_10mm): return T2sum10mmGoodHits[sl];
  }
  return 0;
}



/* --------------- HBOOK Copy and Store for Debugging ------------------
   --------------------------------------------------------------------*/
#ifdef USEHBOOK
int  OtrMonHbookCopyandStore() {
  int i, junk, nb, ids[128];
  printf("OtrMonHbookCopyandStore> Write HBook File otrmon.hbook\n");
  hlimit(PAWC_SIZE);

  /* copy rhp to HBOOK files  */
  if (OtrMonAlgorithm&OTRMONAlgoFed)  OtrMonFedMyHistIds(&nb, ids);
  for (i=0; i<nb; i++) {
    printf("Putting Hist ID %d to file\n",ids[i]); 
    rhp_2hbook(ids[i], junk, "//PAWC");
  }
  if (OtrMonAlgorithm&OTRMONAlgoOccup)  OtrMonOccupMyHistIds(&nb, ids);
  for (i=0; i<nb; i++) {
    printf("Putting Hist ID %d to file\n",ids[i]); 
    rhp_2hbook(ids[i], junk, "//PAWC");
  }
  if (OtrMonAlgorithm&OTRMONAlgoStatus)  OtrMonStatusMyHistIds(&nb, ids);
  for (i=0; i<nb; i++) {
    printf("Putting Hist ID %d to file\n",ids[i]); 
    rhp_2hbook(ids[i], junk, "//PAWC");
  }
  if (OtrMonAlgorithm&OTRMONAlgoSynchr)  OtrMonSynchrMyHistIds(&nb, ids);
  for (i=0; i<nb; i++) {
    printf("Putting Hist ID %d to file\n",ids[i]); 
    rhp_2hbook(ids[i], junk, "//PAWC");
  }
  rhp_2file("otrmon.hbook");
  return OTRIValid;
}
#endif


